# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

#cython: boundscheck=False, wraparound=False

"""\
Low-level Gromacs XDR trajectory reading — :mod:`MDAnalysis.lib.formats.libmdaxdr`
----------------------------------------------------------------------------------

:mod:`libmdaxdr` contains the classes :class:`XTCFile` and
:class:`TRRFile`. Both can be used to read and write frames from and to
Gromacs_ XTC and TRR files. These classes are used internally by MDAnalysis in
:mod:`MDAnalysis.coordinates.XTC` and :mod:`MDAnalysis.coordinates.TRR`. They
behave similar to normal file objects.

For example, one can use a :class:`XTCFile` to directly calculate mean
coordinates (where the coordinates are stored in `x` attribute of the
:class:`namedtuple` `frame`):

.. code-block:: python
   :emphasize-lines: 1,2,5

   with XTCFile("trajectory.xtc") as xtc:
      n_atoms = xtc.n_atoms
      mean = np.zeros((n_atoms, 3))
      # iterate over trajectory
      for frame in xtc:
          mean += frame.x

The :class:`XTCFile` class can be useful as a compressed storage format.

Besides iteration, :class:`XTCFile` and :class:`TRRFile` one can also seek to
arbitrary frames using the :meth:`~XTCFile.seek` method. This is provided by
lazily generating a offset list for stored frames. The offset list is generated
the first time :func:`len` or :`~XTCFile.seek` is called.

(For more details on how to use :class:`XTCFile` and :class:`TRRFile` on their
own please see the source code in `lib/formats/libmdaxdr.pyx`_ for the time being.)


.. _Gromacs: http://www.gromacs.org
.. _`lib/formats/libmdaxdr.pyx`:
   https://github.com/MDAnalysis/mdanalysis/blob/develop/package/MDAnalysis/lib/formats/libmdaxdr.pyx
"""

cimport numpy as cnp
cimport cython
from MDAnalysis.lib.formats.cython_util cimport ptr_to_ndarray
from libc.stdint cimport int64_t

from libc.stdio cimport SEEK_SET, SEEK_CUR, SEEK_END
_whence_vals = {"SEEK_SET": SEEK_SET, "SEEK_CUR": SEEK_CUR, "SEEK_END": SEEK_END}

cdef enum:
    EOK = 0
    EHEADER = 1
    ESTRING = 2
    EDOUBLE = 3
    EINTEGER = 4
    EFLOAT = 5
    EUNSIGNEDINTEGER = 6
    ECOMPRESSION = 7
    ECLOSE = 8
    EMAGIC = 9
    EMEMORY = 10
    EENDOFFILE = 11
    ENOTFOUND = 12

error_message = ['OK', 'header', 'string', 'double', 'integer',
                 'float', 'unsigned integer', 'compression',
                 'close', 'magic', 'memory', 'endoffile',
                 'notfound']

import cython
import numpy as np
from os.path import exists
from collections import namedtuple

cnp.import_array()

ctypedef float DTYPE_T
DTYPE = np.float32
cdef int DIMS = 3
cdef int HASX = 1
cdef int HASV = 2
cdef int HASF = 4


cdef class _XDRFile:
    """Base python wrapper for gromacs-xdr formats

    This class can be similar to the normal file objects in python. The read()
    function will return a frame and all information in it instead of a single
    line. Additionally the context-manager protocoll is supported as well.

    Parameters
    ----------
    fname : str
        The filename to open.
    mode : ('r', 'w')
        The mode in which to open the file, either 'r' read or 'w' write

    Raises
    ------
    IOError

    Note
    ----
    This class can't be initialized use one of the subclasses XTCFile, TRRFile
    """

    def __cinit__(self, fname, mode='r'):
        self.fname = fname.encode('utf-8')
        self.is_open = False
        self.open(self.fname, mode)

    def __dealloc__(self):
        self.close()

    def close(self):
        """Close the open XDR file

        Raises
        ------
        IOError
        """
        res = 1
        if self.is_open:
            res = xdrfile_close(self.xfp)
            self.is_open = False
            if res != EOK:
                raise IOError('Couldn\'t close file: {}  XDR error = {}'.format(
                    self.fname, error_message[res]))
        # forget old offsets in case we open a different file with the same instance.
        self._has_offsets = False
        return True

    def open(self, fname, mode):
        """Open a XDR file

        If another XDR file is currently opened it will be closed

        Parameters
        ----------
        fname : str
            The filename to open.
        mode : ('r', 'w')
            The mode in which to open the file, either 'r' read or 'w' write

        Raises
        ------
        IOError
        """
        if self.is_open:
            self.close()
        self.fname = fname
        self.n_atoms = 0
        self.reached_eof = False
        self.current_frame = 0

        cdef int return_code

        if mode == 'r':
            opening_mode = b'r'
        elif mode == 'w':
            opening_mode = b'w'
        else:
            raise IOError('mode must be one of "r" or "w", you '
                          'supplied {}'.format(mode))
        self.mode = mode

        if self.mode == 'r':
            if not exists(self.fname):
                raise IOError('File does not exist: {}'.format(self.fname))

            return_code, self.n_atoms = self._calc_natoms(fname)

            if return_code != EOK:
                raise IOError('XDR read error = {}'.format(
                    error_message[return_code]))
            if self.n_atoms <= 0:
                raise IOError('Couldn\'t read number of atoms: {}'.format(
                    fname))

        self.xfp = xdrfile_open(fname, opening_mode)
        if self.xfp is NULL:
            raise IOError('Error opening XTC/TRR file: {}'.format(self.fname))
        self.is_open = True

    def __enter__(self):
        """Support context manager"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Support context manager"""
        self.close()
        # always propagate exceptions forward
        return False

    def __iter__(self):
        self.close()
        self.open(self.fname, self.mode)
        return self

    def __next__(self):
        """Return next frame"""
        if self.reached_eof:
            raise StopIteration
        return self.read()

    def __len__(self):
        if not self.is_open:
            raise IOError('No file currently opened')
        return self.offsets.size

    def __reduce__(self):
        """low level pickle function. It gives instructions for unpickling which class
        should be created with arguments to `__cinit__` and the current state

        """
        return (self.__class__, (self.fname.decode(), self.mode),
                self.__getstate__())

    def __getstate__(self):
        """
        current state
        """
        return (self.is_open, self.current_frame, self._offsets,
                self._has_offsets)

    def __setstate__(self, state):
        """
        reset state offsets and current frame
        """
        is_open = state[0]
        # by default the file is open
        if not is_open:
            self.close()
            return

        # set them now to avoid recalculation
        self._offsets = state[2]
        self._has_offsets = state[3]

        # where was I
        current_frame = state[1]
        if current_frame < self.offsets.size:
            self.seek(current_frame)
        elif current_frame == self.offsets.size:
            #  cannot seek to self.offsets.size (a.k.a len(_XDRFile));
            #  instead, we seek to the previous frame and read next. 
            #  which is the state of the file when we need to serialize
            #  at the end of the trajectory.
            self.seek(current_frame - 1)
            _ = self.read()
        else:       # pragma: no cover
            raise RuntimeError("Invalid frame number {} > {} -- this should"
                               "not happen.".format(current_frame,
                                                    self.offsets.size))

    def seek(self, int frame):
        """Seek to Frame.

        Please note that this function will generate internal file offsets if
        they haven't been set before. For large file this means the first seek
        can be very slow. Later seeks will be very fast.

        Parameters
        ----------
        frame : int
            seek the file to given frame

        Raises
        ------
        IOError
        """
        cdef int64_t offset
        if frame == 0:
            offset = 0
        elif frame < 0:
            raise IOError("Can't seek to negative frame")
        else:
            if frame >= self.offsets.size:
                raise EOFError('Trying to seek over max number of frames')
            offset = self.offsets[frame]
        self.reached_eof = False
        ok = xdr_seek(self.xfp, offset, SEEK_SET)
        if ok != EOK:
            # errno is the direct system error, not from the XDR library.
            # errno=22 can be cryptic, since it means a wrong 'whence'
            #  parameter but it will also be issued when the resulting file
            #  offset is negative, or if the offset overflows the filesystem
            #  limit (ext3 is 16TB, for instance).
            raise IOError("XDR seek failed with system errno={}".format(ok))
        self.current_frame = frame

    def _bytes_seek(self, int64_t offset, whence="SEEK_SET"):
        """Low-level access to the stream repositioning xdr_seek call.

        Beware that this function will not update :attr:`current_frame`,
        even on a successful seek, and might render the :class:`Reader`
        unstable.

        Parameters
        ----------
        offset : int
            move the stream *offset* bytes relative to *whence*
            (can be negative).

        whence : str, optional
            must be one of:
             "SEEK_SET": *offset* is applied relative to the beginning
                         of the file.
             "SEEK_CUR": *offset* is applied relative to the current
                         position in the stream.
             "SEEK_END": *offset* is applied relative to the end of
                         the file (a positive *offset* will seek beyond the
                         end of the file, without an error)

        Raises
        ------
        IOError
        """
        cdef int whn
        cdef int64_t offst

        try:
            whn = _whence_vals[whence]
        except KeyError:
            raise IOError("Parameter 'whence' must be "
                          "one of {}".format(tuple(_whence_vals.keys())))
        offst = offset
        self.reached_eof = False
        ok = xdr_seek(self.xfp, offst, whn)
        if ok != EOK:
            # See the comments to seek() for hints on errno meaning.
            raise IOError("XDR seek failed with system errno={}".format(ok))

    @property
    def offsets(self):
        """get byte offsets from current xdr file

        See Also
        --------
        set_offsets
        """
        if not self._has_offsets:
            self._offsets = self.calc_offsets()
            self._has_offsets = True
        return self._offsets

    def set_offsets(self, cnp.ndarray offsets):
        """set frame offsets"""
        self._offsets = offsets
        self._has_offsets = True

    def tell(self):
        """Get current frame"""
        return self.current_frame

    def _bytes_tell(self):
        """Low-level call to xdr_tell to get current byte offset."""
        return xdr_tell(self.xfp)

TRRFrame = namedtuple('TRRFrame', 'x v f box step time lmbda hasx hasv hasf')


cdef class TRRFile(_XDRFile):
    """File-like wrapper for gromacs TRR files.

    This class can be similar to the normal file objects in python. The read()
    function will return a frame and all information in it instead of a single
    line. Additionally the context-manager protocoll is supported as well.

    Parameters
    ----------
    fname : str
        The filename to open.
    mode : ('r', 'w')
        The mode in which to open the file, either 'r' read or 'w' write

    Examples
    --------
    >>> from MDAnalysis.lib.formats.libmdaxdr import TRRFile
    >>> with TRRFile('foo.trr') as f:
    >>>     for frame in f:
    >>>         print(frame.x)

    Notes
    -----
    This class can be pickled. The pickle will store filename, mode, current
    frame and offsets


    .. versionchanged:: 2.4.0
       Added read_direct_xvf method to read into an existing positions array
    """

    def _calc_natoms(self, fname):
        cdef int n_atoms
        cdef int return_code = read_trr_natoms(fname, &n_atoms)
        return return_code, n_atoms

    def calc_offsets(self):
        """read byte offsets from TRR file directly"""
        if not self.is_open:
            return np.array([])
        cdef int n_frames = 0
        cdef int est_nframes = 0
        cdef int64_t* offsets = NULL
        cdef int ok = read_trr_n_frames(self.fname, &n_frames, &est_nframes, &offsets)
        if ok != EOK:
            raise IOError("TRR couldn't calculate offsets. "
                          "XDR error = {}".format(error_message[ok]))
        # the read_xtc_n_frames allocates memory for the offsets with an
        # overestimation. This number is saved in est_nframes and we need to
        # tell the new numpy array about the whole allocated memory to avoid
        # memory leaks.
        cdef cnp.npy_intp[1] dim
        dim[0] = 1
        cdef cnp.ndarray[cnp.int64_t, ndim=1] dims = cnp.PyArray_EMPTY(1, dim, cnp.NPY_INT64, 0)
        dims[0] = est_nframes
        # this handles freeing the allocated memory correctly.
        cdef cnp.ndarray nd_offsets = ptr_to_ndarray(<void*> offsets, dims, cnp.NPY_INT64)
        return nd_offsets[:n_frames]

    def read(self):
        """Read next frame in the TRR file

        Returns
        -------
        frame : libmdaxdr.TRRFrame
            namedtuple with frame information

        See Also
        --------
        TRRFrame
        XTCFile

        Raises
        ------
        IOError
        """
        if self.reached_eof:
            raise EOFError('Reached last frame in TRR, seek to 0')
        if not self.is_open:
            raise IOError('No file opened')
        if self.mode != 'r':
            raise IOError('File opened in mode: {}. Reading only allow '
                               'in mode "r"'.format('self.mode'))

        return_code = 1
        cdef int step = 0
        cdef int has_prop = 0
        cdef float time = 0
        cdef float lmbda = 0

        cdef cnp.npy_intp[2] dim
        dim[0] = self.n_atoms
        dim[1] = DIMS

        cdef cnp.npy_intp[2] unitcell_dim
        unitcell_dim[0] = DIMS
        unitcell_dim[1] = DIMS

        cdef cnp.ndarray[cnp.float32_t, ndim=2] xyz = cnp.PyArray_EMPTY(2, dim, cnp.NPY_FLOAT32, 0)
        cdef cnp.ndarray[cnp.float32_t, ndim=2] velocity = cnp.PyArray_EMPTY(2, dim, cnp.NPY_FLOAT32, 0)
        cdef cnp.ndarray[cnp.float32_t, ndim=2] forces = cnp.PyArray_EMPTY(2, dim, cnp.NPY_FLOAT32, 0)
        cdef cnp.ndarray[cnp.float32_t, ndim=2] box = cnp.PyArray_EMPTY(2, unitcell_dim, cnp.NPY_FLOAT32, 0)

        return_code = read_trr(self.xfp, self.n_atoms, <int*> &step,
                                      &time, &lmbda, <matrix>box.data,
                                      <rvec*>xyz.data,
                                      <rvec*>velocity.data,
                                      <rvec*>forces.data,
                                      <int*> &has_prop)
        # trr are a bit weird. Reading after the last frame always always
        # results in an integer error while reading. I tried it also with trr
        # produced by different codes (Gromacs, ...).
        # In a trr the integer error seems to indicate that the file is ending.
        # There might be corrupted files where this is a legitimate error. But
        # then we just can't read it and stop there which is not too bad.
        if return_code == EENDOFFILE or return_code == EINTEGER:
            self.reached_eof = True
            raise StopIteration

        if return_code != EOK:
            raise IOError('TRR read error = {}'.format(
                error_message[return_code]))

        self.current_frame += 1

        has_x = bool(has_prop & HASX)
        has_v = bool(has_prop & HASV)
        has_f = bool(has_prop & HASF)
        return TRRFrame(xyz, velocity, forces, box, step, time, lmbda,
                        has_x, has_v, has_f)

    def read_direct_xvf(self, cnp.float32_t[:, ::1] positions,
                        cnp.float32_t[:, ::1] velocities,
                        cnp.float32_t[:, ::1] forces,):
        """
        Read next frame in the TRR file with positions read directly into
        a pre-existing array.

        Parameters
        ----------
        positions : np.ndarray
            positions array to read positions into

        Returns
        -------
        frame : libmdaxdr.TRRFrame
            namedtuple with frame information

        See Also
        --------
        TRRFrame
        XTCFile

        Raises
        ------
        IOError


        .. versionadded:: 2.4.0
        """
        if self.reached_eof:
            raise EOFError('Reached last frame in TRR, seek to 0')
        if not self.is_open:
            raise IOError('No file opened')
        if self.mode != 'r':
            raise IOError('File opened in mode: {}. Reading only allow '
                               'in mode "r"'.format('self.mode'))

        cdef int return_code = 1
        cdef int step = 0
        cdef int has_prop = 0
        cdef float time = 0
        cdef float lmbda = 0

        cdef cnp.npy_intp[2] unitcell_dim
        unitcell_dim[0] = DIMS
        unitcell_dim[1] = DIMS


        cdef cnp.ndarray[cnp.float32_t, ndim=2] box = cnp.PyArray_EMPTY(2, unitcell_dim, cnp.NPY_FLOAT32, 0)

        return_code = read_trr(self.xfp, self.n_atoms, <int*> &step,
                                      &time, &lmbda, <matrix>box.data,
                                      <rvec*>&positions[0,0],
                                      <rvec*>&velocities[0,0],
                                      <rvec*>&forces[0,0],
                                      <int*> &has_prop)
        # trr are a bit weird. Reading after the last frame always always
        # results in an integer error while reading. I tried it also with trr
        # produced by different codes (Gromacs, ...).
        # In a trr the integer error seems to indicate that the file is ending.
        # There might be corrupted files where this is a legitimate error. But
        # then we just can't read it and stop there which is not too bad.
        if return_code == EENDOFFILE or return_code == EINTEGER:
            self.reached_eof = True
            raise StopIteration

        if return_code != EOK:
            raise IOError('TRR read error = {}'.format(error_message[return_code]))

        self.current_frame += 1

        has_x = bool(has_prop & HASX)
        has_v = bool(has_prop & HASV)
        has_f = bool(has_prop & HASF)
        return TRRFrame(positions, velocities, forces, box, step, time, lmbda,
                        has_x, has_v, has_f)

    def write(self, xyz, velocity, forces, box, int step, float time,
              float _lambda, int natoms):
        """write one frame into TRR file.

        Parameters
        ----------
        xyz : array_like, shape=(`n_atoms`, 3), optional
            cartesion coordinates. Only written if not ``None``.
        velocity : array_like, shape=(`n_atoms`, 3), optional
            cartesion velocities. Only written if not ``None``.
        forces : array_like, shape=(`n_atoms`, 3), optional
            cartesion forces. Only written if not ``None``.
        box : array_like, shape=(3, 3)
            Box vectors for this frame
        step : int
            current step number, 1 indexed
        time : float
            current time
        _lambda : float
            current lambda value
        natoms : int
            number of atoms in frame

        Raises
        ------
        IOError
        """
        if self.mode != 'w' :
            raise IOError('File opened in mode: {}. Writing only allow '
                          'in mode "w"'.format('self.mode'))

        cdef float* xyz_ptr = NULL
        cdef float* velocity_ptr = NULL
        cdef float* forces_ptr = NULL

        # if any of the xyz/velocity/forces contain any data we use the helper
        # defined here to get a pointer to their first element. This is the only
        # way I know to have a nice pythonic API to the function that can accept
        # array-like inputs or things like None.
        cdef cnp.ndarray xyz_helper
        cdef cnp.ndarray velocity_helper
        cdef cnp.ndarray forces_helper

        if xyz is not None:
            xyz = np.asarray(xyz)
            xyz_helper = np.ascontiguousarray(xyz, dtype=DTYPE)
            xyz_ptr = <float*>xyz_helper.data
        if velocity is not None:
            velocity = np.asarray(velocity)
            velocity_helper = np.ascontiguousarray(velocity, dtype=DTYPE)
            velocity_ptr = <float*>velocity_helper.data
        if forces is not None:
            forces = np.asarray(forces)
            forces_helper = np.ascontiguousarray(forces, dtype=DTYPE)
            forces_ptr = <float*>forces_helper.data

        box = np.asarray(box)
        cdef cnp.ndarray box_helper = np.ascontiguousarray(box, dtype=DTYPE)
        cdef float* box_ptr = <float*>box_helper.data

        if self.current_frame == 0:
            self.box = box
            self.n_atoms = natoms
        else:
            if self.n_atoms != natoms:
                raise IOError('Previous frames contained {} atoms. You '
                              'are trying to write {} atoms.'.format(
                                  self.n_atoms, natoms))
            if xyz is not None and self.n_atoms != xyz.shape[0]:
                raise IOError('Previous frames xyz contained {} atoms. You '
                              'are trying to write {} atoms.'.format(
                                  self.n_atoms, xyz.shape[0]))
            if velocity is not None and self.n_atoms != velocity.shape[0]:
                raise IOError('Previous frames velocity contained {} atoms. You '
                              'are trying to write {} atoms.'.format(
                                  self.n_atoms, velocity.shape[0]))
            if forces is not None and self.n_atoms != forces.shape[0]:
                raise IOError('Previous frames forces contained {} atoms. You '
                              'are trying to write {} atoms.'.format(
                                  self.n_atoms, forces.shape[0]))

        return_code = write_trr(self.xfp, self.n_atoms, step, time,
                                _lambda, <matrix> box_ptr,
                                <rvec*> xyz_ptr,
                                <rvec*> velocity_ptr,
                                <rvec*> forces_ptr)
        if return_code != EOK:
            raise IOError('TRR write error = {}'.format(
                error_message[return_code]))

        self.current_frame += 1


XTCFrame = namedtuple('XTCFrame', 'x box step time prec')


cdef class XTCFile(_XDRFile):
    """File-like wrapper for gromacs XTC files.

    This class can be similar to the normal file objects in python. The read()
    function will return a frame and all information in it instead of a single
    line. Additionally the context-manager protocoll is supported as well.

    Parameters
    ----------
    fname : str
        The filename to open.
    mode : ('r', 'w')
        The mode in which to open the file, either 'r' read or 'w' write

    Examples
    --------
    >>> from MDAnalysis.lib.formats.libmdaxdr import XTCFile
    >>> with XTCFile('foo.trr') as f:
    >>>     for frame in f:
    >>>         print(frame.x)

    Notes
    -----
    This class can be pickled. The pickle will store filename, mode, current
    frame and offsets

    .. versionchanged:: 2.4.0
       Added read_direct_x method to read into an existing positions array
    """

    def _calc_natoms(self, fname):
        cdef int n_atoms
        cdef int return_code = read_xtc_natoms(fname, &n_atoms)
        return return_code, n_atoms

    def calc_offsets(self):
        """Calculate offsets from XTC file directly"""
        if not self.is_open:
            return np.array([])
        cdef int n_frames = 0
        cdef int est_nframes = 0
        cdef int64_t* offsets = NULL
        cdef int ok = read_xtc_n_frames(self.fname, &n_frames, &est_nframes, &offsets)
        if ok != EOK:
            raise IOError("XTC couldn't calculate offsets. "
                          "XDR error = {}".format(error_message[ok]))
        # the read_xtc_n_frames allocates memory for the offsets with an
        # overestimation. This number is saved in est_nframes and we need to
        # tell the new numpy array about the whole allocated memory to avoid
        # memory leaks.
        cdef cnp.npy_intp[1] dim
        dim[0] = 1
        cdef cnp.ndarray[cnp.int64_t, ndim=1] dims = cnp.PyArray_EMPTY(1, dim, cnp.NPY_INT64, 0)
        dims[0] = est_nframes
        # this handles freeing the allocated memory correctly.
        cdef cnp.ndarray nd_offsets = ptr_to_ndarray(<void*> offsets, dims, cnp.NPY_INT64)
        return nd_offsets[:n_frames]

    def read(self):
        """Read next frame in the XTC file

        Returns
        -------
        frame : libmdaxdr.XTCFrame
            namedtuple with frame information

        See Also
        --------
        XTCFrame
        TRRFile

        Raises
        ------
        IOError
        """
        if self.reached_eof:
            raise EOFError('Reached last frame in XTC, seek to 0')
        if not self.is_open:
            raise IOError('No file opened')
        if self.mode != 'r':
            raise IOError('File opened in mode: {}. Reading only allow '
                               'in mode "r"'.format('self.mode'))

        return_code = 1
        cdef int step
        cdef float time, prec

        cdef cnp.npy_intp[2] dim
        dim[0] = self.n_atoms
        dim[1] = DIMS

        cdef cnp.npy_intp[2] unitcell_dim
        unitcell_dim[0] = DIMS
        unitcell_dim[1] = DIMS

        cdef cnp.ndarray[cnp.float32_t, ndim=2] xyz = cnp.PyArray_EMPTY(2, dim, cnp.NPY_FLOAT32, 0)
        cdef cnp.ndarray[cnp.float32_t, ndim=2] box = cnp.PyArray_EMPTY(2, unitcell_dim, cnp.NPY_FLOAT32, 0)

        return_code = read_xtc(self.xfp, self.n_atoms, <int*> &step,
                                      &time, <matrix>box.data,
                                      <rvec*>xyz.data, <float*> &prec)

        if return_code == EENDOFFILE:
            self.reached_eof = True
            raise StopIteration

        if return_code != EOK:
            raise IOError('XTC read error = {}'.format(error_message[return_code]))
        self.current_frame += 1

        return XTCFrame(xyz, box, step, time, prec)

    def read_direct_x(self, cnp.float32_t[:, ::1] positions):
        """
        Read next frame in the XTC file with positions read directly into
        a pre-existing array.

        Parameters
        ----------
        positions : np.ndarray
           positions array to read positions into

        Returns
        -------
        frame : libmdaxdr.XTCFrame
            namedtuple with frame information

        See Also
        --------
        XTCFrame
        TRRFile

        Raises
        ------
        IOError


        .. versionadded:: 2.4.0
        """
        if self.reached_eof:
            raise EOFError('Reached last frame in XTC, seek to 0')
        if not self.is_open:
            raise IOError('No file opened')
        if self.mode != 'r':
            raise IOError('File opened in mode: {}. Reading only allow '
                               'in mode "r"'.format('self.mode'))

        return_code = 1
        cdef int step
        cdef float time, prec
        cdef cnp.npy_intp[2] unitcell_dim
        unitcell_dim[0] = DIMS
        unitcell_dim[1] = DIMS

        cdef cnp.ndarray[cnp.float32_t, ndim=2] box = cnp.PyArray_EMPTY(2, unitcell_dim, cnp.NPY_FLOAT32, 0)


        return_code = read_xtc(self.xfp, self.n_atoms, <int*> &step,
                                      &time, <matrix>box.data,
                                      <rvec*>&positions[0,0], <float*> &prec)

        if return_code == EENDOFFILE:
            self.reached_eof = True
            raise StopIteration

        if return_code != EOK:
            raise IOError('XTC read error = {}'.format(error_message[return_code]))
        self.current_frame += 1

        return  XTCFrame(positions, box, step, time, prec)


    def write(self, xyz, box, int step, float time, float precision=1000):
        """write one frame to the XTC file

        Parameters
        ----------
        xyz : array_like, shape=(`n_atoms`, 3)
            cartesion coordinates
        box : array_like, shape=(3, 3)
            Box vectors for this frame
        step : int
            current step number, 1 indexed
        time : float
            current time
        precision : float (optional)
            precision of saved trajectory, see Notes


        Notes
        -----
        Gromacs specifies the precision a little bit strange. The coordinates
        will be multiplied by precision and then converted to an integer. This
        means a precision of 1000 yields an accuracy of 3 decimal places.

        Raises
        ------
        IOError

        """
        if self.mode != 'w':
            raise IOError('File opened in mode: {}. Writing only allow '
                          'in mode "w"'.format('self.mode'))

        xyz = np.asarray(xyz, dtype=np.float32)
        box = np.asarray(box, dtype=np.float32)

        cdef DTYPE_T[:, ::1] xyz_view = cnp.PyArray_GETCONTIGUOUS(xyz)
        cdef DTYPE_T[:, ::1] box_view = cnp.PyArray_GETCONTIGUOUS(box)

        if self.current_frame == 0:
            self.n_atoms = xyz.shape[0]
            self.box = box
            self.precision = precision
        else:
            if self.n_atoms != xyz.shape[0]:
                raise IOError('Previous frames contained {} atoms. You '
                              'are trying to write {} atoms.'.format(
                                  self.n_atoms, xyz.shape[1]))
            if self.precision != precision:
                raise IOError('Previous frames used precision of {}. You '
                              'are trying to use {}'.format(
                                  self.precision, precision))

        return_code = write_xtc(self.xfp, self.n_atoms, step, time,
                                       <matrix>&box_view[0, 0],
                                       <rvec*>&xyz_view[0, 0], precision)
        if return_code != EOK:
            raise IOError('XTC write error = {}'.format(
                error_message[return_code]))

        self.current_frame += 1



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

r"""
ITP topology parser
===================

Reads a GROMACS ITP_ or TOP_ file to build the system. The topology will
contain atom IDs, segids, residue IDs, residue names, atom names, atom types,
charges, chargegroups, masses, moltypes, and molnums.
Any masses that are in the file will be read; any missing values will be guessed.
Bonds, angles, dihedrals and impropers are also read from the file.

If an ITP file is passed without a ``[ molecules ]`` directive, passing 
``infer_system=True`` (the default option) will create a Universe with 
1 molecule of each defined ``moleculetype``. 
If a ``[ molecules ]`` section is present, ``infer_system`` is ignored.

If files are included with the `#include` directive, they will also be read.
If they are not in the working directory, ITPParser will look for them in the
``include_dir`` directory. By default, this is set to
``include_dir="/usr/local/gromacs/share/gromacs/top/"``.
Variables can be defined with the `#define` directive in files, or by passing
in :ref:`keyword arguments <itp-define-kwargs>`.

Examples
--------

.. code-block:: python

    import MDAnalysis as mda
    from MDAnalysis.tests.datafiles import ITP_tip5p

    #  override charge of HW2 atom (defined in file as HW2_CHARGE)
    u = mda.Universe(ITP_tip5p, HW2_CHARGE=2, infer_system=True)

.. note::

    AMBER also uses topology files with the .top extension. To use ITPParser
    to read GROMACS top files, pass ``topology_format='ITP'``.

    ::

        import MDAnalysis as mda
        u = mda.Universe('topol.top', topology_format='ITP')


.. _itp-define-kwargs:

Preprocessor variables
----------------------

ITP files are often defined with lines that depend on 
whether a keyword flag is given. For example, this modified TIP5P water file:

.. code-block:: none

    [ moleculetype ]
    ; molname       nrexcl
    SOL             2

    #ifndef HW1_CHARGE
        #define HW1_CHARGE 0.241
    #endif

    #define HW2_CHARGE 0.241

    [ atoms ]
    ; id    at type res nr  residu name     at name         cg nr   charge
    1       opls_118     1       SOL              OW             1       0
    2       opls_119     1       SOL             HW1             1       HW1_CHARGE
    3       opls_119     1       SOL             HW2             1       HW2_CHARGE
    4       opls_120     1       SOL             LP1             1      -0.241
    5       opls_120     1       SOL             LP2             1      -0.241
    #ifdef EXTRA_ATOMS  ; added for keyword tests
    6       opls_120     1       SOL             LP3             1      -0.241
    7       opls_120     1       SOL             LP4             1       0.241
    #endif


Define these preprocessor variables by passing keyword arguments. Any arguments that you 
pass in *override* any variables defined in the file. For example, the universe below 
will have charges of 3 for the HW1 and HW2 atoms::

    import MDAnalysis as mda
    from MDAnalysis.tests.datafiles import ITP_tip5p

    u = mda.Universe(ITP_tip5p, EXTRA_ATOMS=True, HW1_CHARGE=3, HW2_CHARGE=3)

These keyword variables are **case-sensitive**. Note that if you set keywords to
``False`` or ``None``, they will be treated as if they are not defined in #ifdef conditions.

For example, the universe below will only have 5 atoms. ::

    u = mda.Universe(ITP_tip5p, EXTRA_ATOMS=False)


.. _ITP: http://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html#molecule-itp-file

.. _TOP: http://manual.gromacs.org/current/reference-manual/file-formats.html#top

Classes
-------

.. autoclass:: ITPParser
   :members:
   :inherited-members:

"""
from collections import defaultdict
import os

import logging
import numpy as np

import warnings
from ..lib.util import openany
from ..guesser.tables import SYMB2Z
from ..guesser.default_guesser import DefaultGuesser
from .base import TopologyReaderBase, change_squash, reduce_singular
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Atomtypes,
    Masses,
    Moltypes,
    Molnums,
    Charges,
    Resids,
    Resnums,
    Resnames,
    Segids,
    Bonds,
    Angles,
    Dihedrals,
    Impropers,
    AtomAttr,
    Elements,
)
from ..core.topology import Topology


class Chargegroups(AtomAttr):
    """The charge group for each Atom"""

    attrname = "chargegroups"
    singular = "chargegroup"


class GmxTopIterator:
    """
    Iterate over the lines of a TOP/ITP file and its included files

    The iterator strips comments, deals with #ifdef and #ifndef conditions,
    and substitutes defined variables.

    Defined variables passed into ``__init__`` *override* anything defined in the file.
    """

    def __init__(self, path, include_dir, defines):
        self._original_defines = defines
        self.defines = dict(**defines)  # copy
        self.include_dir = include_dir
        self.file_stack = [path]
        self.starting_file = path

    @property
    def current_file(self):
        return self.file_stack[-1]

    def __iter__(self):
        for line in self.iter_from_file(self.starting_file):
            yield line

    def iter_from_file(self, path):
        found_path = self.find_path(path)
        with openany(found_path) as infile:
            self.file_stack.append(infile)

            for line in self.clean_file_lines(infile):
                if line.startswith("#include"):
                    inc = line.split(None, 1)[1][1:-1]
                    for line in self.iter_from_file(inc):
                        yield line
                elif line.startswith("#define"):
                    self.define(line)
                elif line.startswith("#if"):
                    self.do_if(line, infile)
                elif line.startswith("#else"):
                    self.skip_until_endif(infile)
                elif line.startswith("#"):  # ignore #if and others
                    pass
                elif line:
                    line = self.substitute_defined(line)
                    yield line
            self.file_stack.pop()

    def define(self, line):
        try:
            _, variable, value = line.split(None, 2)
        except ValueError:
            _, variable = line.split()
            value = True

        # kwargs overrides files
        if variable not in self._original_defines:
            self.defines[variable] = value

    def substitute_defined(self, line):
        split = line.split()
        for k, v in self.defines.items():
            if k in split:
                split[split.index(k)] = str(v)
        line = " ".join(split)
        return line

    def clean_file_lines(self, infile):
        for line in infile:
            line = line.split(";")[0].strip()  # ; is for comments
            yield line

    def do_if(self, line, infile):
        ifdef, variable = line.split()
        if ifdef == "#ifdef":
            if self.defines.get(variable) in (False, None):
                self.skip_until_else(infile)
        elif ifdef == "#ifndef":
            if self.defines.get(variable) not in (False, None):
                self.skip_until_else(infile)

    def skip_until_else(self, infile):
        """Skip lines until #if condition ends"""
        for line in self.clean_file_lines(infile):
            if line.startswith("#if"):
                self.skip_until_endif(infile)
            elif line.startswith("#endif") or line.startswith("#else"):
                break
        else:
            raise IOError("Missing #endif in {}".format(self.current_file))

    def skip_until_endif(self, infile):
        """Skip lines until #endif"""
        for line in self.clean_file_lines(infile):
            if line.startswith("#if"):
                self.skip_until_endif(infile)
            elif line.startswith("#endif"):
                break
        else:
            raise IOError("Missing #endif in {}".format(self.current_file))

    def find_path(self, path):
        try:
            # in case of TextIOWrapper
            current_file = self.current_file.name
        except AttributeError:
            current_file = self.current_file

        try:
            path = os.path.abspath(path.name)
        except AttributeError:
            pass
        current_dir = os.path.dirname(current_file)
        dir_path = os.path.join(current_dir, path)
        if os.path.exists(dir_path):
            return dir_path
        include_path = os.path.join(self.include_dir, path)
        if os.path.exists(include_path):
            return include_path
        raise IOError("Could not find {}".format(path))


class Molecule:
    """Store moleculetype-specific attributes"""

    def __init__(self, name):
        self.name = name
        self.ids = []
        self.types = []
        self.resids = []
        self.resnames = []
        self.names = []
        self.chargegroups = []
        self.charges = []
        self.masses = []

        self.bonds = defaultdict(list)
        self.angles = defaultdict(list)
        self.dihedrals = defaultdict(list)
        self.impropers = defaultdict(list)

        self.parsers = {
            "atoms": self.parse_atoms,
            "bonds": self.parse_bonds,
            "angles": self.parse_angles,
            "dihedrals": self.parse_dihedrals,
            "constraints": self.parse_constraints,
            "settles": self.parse_settles,
        }

        self.resolved_residue_attrs = False

    @property
    def atom_order(self):
        return [
            self.ids,
            self.types,
            self.resids,
            self.resnames,
            self.names,
            self.chargegroups,
            self.charges,
            self.masses,
        ]

    @property
    def params(self):
        return [self.bonds, self.angles, self.dihedrals, self.impropers]

    def parse_atoms(self, line):
        values = line.split()
        for lst in self.atom_order:
            try:
                lst.append(values.pop(0))
            except IndexError:  # ran out of values
                lst.append("")

    def parse_bonds(self, line):
        self.add_param(
            line,
            self.bonds,
            n_funct=2,
            funct_values=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
        )

    def parse_angles(self, line):
        self.add_param(
            line,
            self.angles,
            n_funct=3,
            funct_values=(1, 2, 3, 4, 5, 6, 8, 10),
        )

    def parse_dihedrals(self, line):
        dih = self.add_param(
            line,
            self.dihedrals,
            n_funct=4,
            funct_values=(1, 3, 5, 8, 9, 10, 11),
        )
        if not dih:
            self.add_param(
                line, self.impropers, n_funct=4, funct_values=(2, 4)
            )

    def parse_constraints(self, line):
        self.add_param(line, self.bonds, n_funct=2, funct_values=(1, 2))

    def parse_settles(self, line):
        # [ settles ] is a triangular constraint for
        # water molecules.
        # In ITP files this is defined with only the
        # oxygen atom index. The next two atoms are
        # assumed to be hydrogens. Unlike TPRParser,
        # the manual only lists this format (as of 2019).
        # These are treated as 2 bonds.
        # No angle component is included to avoid discrepancies
        # with water molecules loaded from different MD engines.
        oxygen, funct, doh, dhh = line.split()
        try:
            base = self.index_ids([oxygen])[0]
        except ValueError:
            pass
        else:
            self.bonds[(base, base + 1)].append("settles")
            self.bonds[(base, base + 2)].append("settles")

    def resolve_residue_attrs(self):
        """Figure out residue borders and assign moltypes and molnums"""
        resids = np.array(self.resids, dtype=np.int32)
        resnames = np.array(self.resnames, dtype=object)
        self.residx, (self.resids, resnames) = change_squash(
            (resids,), (resids, resnames)
        )
        self.resnames = list(resnames)
        self.moltypes = [self.name] * len(self.resids)
        self.molnums = np.array([1] * len(self.resids))

        self.resolved_residue_attrs = True

    def shift_indices(
        self, atomid=0, resid=0, molnum=0, cgnr=0, n_res=0, n_atoms=0
    ):
        """
        Get attributes ready for adding onto a larger topology.

        Shifts atom indices, residue indices, molnums, and chargegroup numbers.

        Returns
        -------
        atom_attrs: list of lists
            attributes in the [ atoms ] section

        new_params: list of dicts
            Bonds, angles, dihedrals, impropers as dicts of shape {indices: parameters}

        molnums: list
        moltypes: list
        residx: list
        """
        if not self.resolved_residue_attrs:
            self.resolve_residue_attrs()

        resids = list(np.array(self.resids) + resid)
        residx = list(np.array(self.residx) + n_res)
        molnums = list(np.array(self.molnums) + molnum)
        ids = list(np.array(self.ids, dtype=int) + atomid)

        try:
            cg = np.array(self.chargegroups, dtype=int)
        except ValueError:
            cg = np.arange(1, len(self.chargegroups) + 1)
        chargegroups = list(cg + cgnr)

        atom_order = [
            ids,
            self.types,
            resids,
            self.resnames,
            self.names,
            chargegroups,
            self.charges,
            self.masses,
        ]

        new_params = []
        for p in self.params:
            new = {}
            for indices, values in p.items():
                new[tuple(np.array(indices) + n_atoms)] = values
            new_params.append(new)

        return atom_order, new_params, molnums, self.moltypes, residx

    def add_param(self, line, container, n_funct=2, funct_values=None):
        """Add defined GROMACS directive lines, only if the funct is in ``funct_values``"""
        if funct_values is None:
            funct_values = []
        values = line.split()
        funct = int(values[n_funct])
        if funct in funct_values:
            try:
                ids = self.index_ids(values[:n_funct])
                container[ids].append(funct)
            except ValueError:
                pass
            return True
        else:
            return False

    def index_ids(self, values):
        """
        Get indices of atom ids (list of strings)
        """
        return tuple(map(self.ids.index, values))


class ITPParser(TopologyReaderBase):
    """Read topology information from a GROMACS ITP_ or TOP_ file.

    Creates a Topology with the following Attributes:
    - ids
    - names
    - types
    - masses
    - charges
    - chargegroups
    - resids
    - resnames
    - segids
    - moltypes
    - molnums
    - bonds
    - angles
    - dihedrals
    - impropers

    .. _ITP: http://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html#molecule-itp-file
    .. _TOP: http://manual.gromacs.org/current/reference-manual/file-formats.html#top

    .. note::

        By default, atomtypes and masses will be guessed on Universe creation
        if they are not read from the input file.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

    .. versionchanged:: 2.2.0
      no longer adds angles for water molecules with SETTLE constraint
    .. versionchanged:: 2.8.0
      Removed mass guessing (attributes guessing takes place now
      through universe.guess_TopologyAttrs() API).
      Added guessed elements if all elements are valid to preserve partial
      mass guessing behavior

    """

    format = "ITP"

    def parse(
        self,
        include_dir="/usr/local/gromacs/share/gromacs/top/",
        infer_system=True,
        **kwargs,
    ):
        """Parse ITP file into Topology

        Parameters
        ----------
        include_dir: str, optional
            A directory in which to look for other files included
            from the original file, if the files are not first found
            in the current directory.
            Default: "/usr/local/gromacs/share/gromacs/top/"

        infer_system: bool, optional (default True)
            If a ``[ molecules ]`` directive is not found within the the
            topology file, create a Topology with one of every
            ``[ moleculetype ]`` defined. If a ``[ molecules ]`` directive is
            found, this keyword is ignored.

        Returns
        -------
        MDAnalysis *Topology* object
        """

        self.atomtypes = {}
        self.molecules = {}
        self._molecules = []  # for order
        self.current_mol = None
        self.parser = self._pass
        self.system_molecules = []

        # Open and check itp validity
        with openany(self.filename) as itpfile:
            self.lines = GmxTopIterator(itpfile, include_dir, kwargs)
            for line in self.lines:
                if "[" in line and "]" in line:
                    section = line.split("[")[1].split("]")[0].strip()

                    if section == "atomtypes":
                        self.parser = self.parse_atomtypes

                    elif section == "moleculetype":
                        self.parser = self.parse_moleculetype

                    elif section == "molecules":
                        self.parser = self.parse_molecules

                    elif self.current_mol:
                        self.parser = self.current_mol.parsers.get(
                            section, self._pass
                        )

                    else:
                        self.parser = self._pass

                else:
                    self.parser(line)

        if not self.system_molecules and infer_system:
            self.system_molecules = [x.name for x in self._molecules]

        self.build_system()

        self.types = np.array(self.types)
        self.charges = np.array(self.charges, dtype=object)
        self.masses = np.array(self.masses, dtype=object)

        if not all(self.charges):
            empty = self.charges == ""
            self.charges[empty] = [
                (
                    self.atomtypes.get(x)["charge"]
                    if x in self.atomtypes.keys()
                    else ""
                )
                for x in self.types[empty]
            ]

        if not all(self.masses):
            empty = self.masses == ""
            self.masses[empty] = [
                (
                    self.atomtypes.get(x)["mass"]
                    if x in self.atomtypes.keys()
                    else ""
                )
                for x in self.types[empty]
            ]

        attrs = []
        # atom stuff
        for vals, Attr, dtype in (
            (self.ids, Atomids, np.int32),
            (self.types, Atomtypes, object),
            (self.names, Atomnames, object),
            (self.chargegroups, Chargegroups, np.int32),
            (self.charges, Charges, np.float32),
        ):
            if all(vals):
                attrs.append(Attr(np.array(vals, dtype=dtype)))

        if not all(self.masses):
            empty = self.masses == ""
            self.masses[empty] = Masses.missing_value_label

        attrs.append(
            Masses(np.array(self.masses, dtype=np.float64), guessed=False)
        )

        self.elements = DefaultGuesser(None).guess_types(self.types)
        if all(e.capitalize() in SYMB2Z for e in self.elements):
            attrs.append(
                Elements(np.array(self.elements, dtype=object), guessed=True)
            )
            warnings.warn(
                "The elements attribute has been populated by guessing "
                "elements from atom types. This behaviour has been "
                "temporarily added to the ITPParser as we transition "
                "to the new guessing API. "
                "This behavior will be removed in release 3.0. "
                "Please see issue #4698 for more information. ",
                DeprecationWarning,
            )
        else:
            warnings.warn(
                "Element information is missing, elements attribute "
                "will not be populated. If needed these can be "
                "guessed using universe.guess_TopologyAttrs("
                "to_guess=['elements'])."
            )

        # residue stuff
        resids = np.array(self.resids, dtype=np.int32)
        resnames = np.array(self.resnames, dtype=object)
        molnums = np.array(self.molnums, dtype=np.int32)
        attrs.append(Resids(resids))
        attrs.append(Resnums(resids.copy()))
        attrs.append(Resnames(resnames))
        attrs.append(Moltypes(np.array(self.moltypes, dtype=object)))
        attrs.append(Molnums(molnums))

        n_atoms = len(self.ids)
        n_residues = len(self.resids)
        n_segments = len(self.system_molecules)
        attrs.append(Segids(np.array(self.system_molecules, dtype=object)))
        segidx = molnums - 1

        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            attrs=attrs,
            atom_resindex=self.residx,
            residue_segindex=segidx,
        )

        # connectivity stuff
        for dct, Attr, attrname in (
            (self.bonds, Bonds, "bonds"),
            (self.angles, Angles, "angles"),
            (self.dihedrals, Dihedrals, "dihedrals"),
            (self.impropers, Impropers, "impropers"),
        ):
            if dct:
                indices, types = zip(*list(dct.items()))
            else:
                indices, types = [], []

            types = [reduce_singular(t) for t in types]

            tattr = Attr(indices, types=types)
            top.add_TopologyAttr(tattr)

        return top

    def _pass(self, line):
        pass

    def parse_atomtypes(self, line):
        keys = ["type_bonded", "atomic_number", "mass", "charge", "p_type"]
        fields = line.split()
        if len(fields[5]) == 1 and fields[5].isalpha():
            values = fields[1:6]
        elif len(fields[3]) == 1 and fields[3].isalpha():
            values = "", "", fields[1], fields[2], fields[3]
        elif len(fields[4]) == 1 and fields[4].isalpha():
            if fields[1][0].isalpha():
                values = fields[1], "", fields[2], fields[3], fields[4]
            else:
                values = "", fields[1], fields[2], fields[3], fields[4]
        self.atomtypes[fields[0]] = dict(zip(keys, values))

    def parse_moleculetype(self, line):
        name = line.split()[0]
        self.current_mol = self.molecules[name] = Molecule(name)
        self._molecules.append(self.current_mol)

    def parse_molecules(self, line):
        name, n_mol = line.split()
        self.system_molecules.extend([name] * int(n_mol))

    def build_system(self):
        self.ids = []
        self.types = []
        self.resids = []
        self.resnames = []
        self.names = []
        self.chargegroups = []
        self.charges = []
        self.masses = []
        self.moltypes = []
        self.molnums = []
        self.residx = []

        self.atom_order = [
            self.ids,
            self.types,
            self.resids,
            self.resnames,
            self.names,
            self.chargegroups,
            self.charges,
            self.masses,
        ]

        self.bonds = defaultdict(list)
        self.angles = defaultdict(list)
        self.dihedrals = defaultdict(list)
        self.impropers = defaultdict(list)

        self.params = [self.bonds, self.angles, self.dihedrals, self.impropers]

        for i, moltype in enumerate(self.system_molecules):
            mol = self.molecules[moltype]

            atomid = self.ids[-1] if self.ids else 0
            resid = self.resids[-1] if self.resids else 0
            cgnr = self.chargegroups[-1] if self.chargegroups else 0
            n_res = len(self.resids)
            n_atoms = len(self.ids)

            shifted = mol.shift_indices(
                atomid=atomid,
                resid=resid,
                n_res=n_res,
                cgnr=cgnr,
                molnum=i,
                n_atoms=n_atoms,
            )
            atom_order, params, molnums, moltypes, residx = shifted

            for system_attr, mol_attr in zip(self.atom_order, atom_order):
                system_attr.extend(mol_attr)

            self.moltypes.extend(moltypes)
            self.molnums.extend(molnums)
            self.residx.extend(residx)

            for system_param, mol_param in zip(self.params, params):
                system_param.update(mol_param)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import pytest
from numpy.testing import assert_equal

import MDAnalysis as mda
from MDAnalysis.lib import NeighborSearch

from MDAnalysisTests.datafiles import PSF, DCD


@pytest.fixture
def universe():
    u = mda.Universe(PSF, DCD)
    u.atoms.translate([100, 100, 100])
    u.atoms.dimensions = [200, 200, 200, 90, 90, 90]
    return u


def test_search(universe):
    """simply check that for a centered protein in a large box periodic
    and non-periodic return the same result"""
    ns = NeighborSearch.AtomNeighborSearch(universe.atoms)
    pns = NeighborSearch.AtomNeighborSearch(
        universe.atoms, universe.atoms.dimensions
    )

    ns_res = ns.search(universe.atoms[20], 20)
    pns_res = pns.search(universe.atoms[20], 20)
    assert_equal(ns_res, pns_res)


def test_zero(universe):
    """Check if empty atomgroup, residue, segments are returned"""
    ns = NeighborSearch.AtomNeighborSearch(universe.atoms[:10])
    ns_res = ns.search(universe.atoms[20], 0.1, level="A")
    assert ns_res == universe.atoms[[]]
    ns_res = ns.search(universe.atoms[20], 0.1, level="R")
    assert ns_res == universe.atoms[[]].residues
    ns_res = ns.search(universe.atoms[20], 0.1, level="S")
    assert ns_res == universe.atoms[[]].segments



import MDAnalysis

try:
    from MDAnalysisTests.datafiles import GRO
except:
    pass


class SimpleSelectionBench(object):
    """Benchmarks for the various MDAnalysis
    simple selection strings.
    """

    params = (
        "protein",
        "backbone",
        "nucleic",
        "nucleicbackbone",
        "resid 1:10",
        "resnum 1:10",
        "resname LYS",
        "name CA",
        "bynum 0:10",
    )

    param_names = ["selection_string"]

    def setup(self, selection_string):
        self.u = MDAnalysis.Universe(GRO)

    def time_simple_selections(self, selection_string):
        """Benchmark simple selections on the protein-based
        standard test GRO file.
        """
        if hasattr(MDAnalysis.Universe, "select_atoms"):
            self.u.select_atoms(selection_string)
        else:
            self.u.selectAtoms(selection_string)


class GeoSelectionBench(object):
    """Benchmarks for the various MDAnalysis
    geometric selection strings.
    """

    # all selection strings verified
    # to produce non-zero atom groups
    # with GRO test file
    params = (
        [
            "around 5.0 resid 1",
            "sphlayer 2.4 6.0 (protein)",
            "sphzone 6.0 (protein)",
            "cylayer 5 10 10 -8 protein",
            "cyzone 15 4 -8 protein",
            "point 5.0 5.0 5.0 3.5",
            "prop z >= 5.0",
            "prop abs z <= 5.0",
        ],
        [True, False],  # updating flags
        [[False, True], [True, False]],
    )  # periodic flags

    # benchmarks should include static &
    # dynamic selections & periodic
    # vs non-periodic
    param_names = ["selection_string", "dynamic_selection", "periodic_selection"]

    def setup(self, selection_string, dynamic_selection, periodic_selection):
        self.u = MDAnalysis.Universe(GRO)

    def time_geometric_selections(
        self, selection_string, dynamic_selection, periodic_selection
    ):

        # TODO: Do we need a kwarg similar to old `use_KDTree_routines`
        # flag? We used to benchmark that.
        self.u.select_atoms(
            selection_string,
            updating=dynamic_selection,
            periodic=periodic_selection[0],
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
PDBQT structure files in MDAnalysis --- :mod:`MDAnalysis.coordinates.PDBQT`
===========================================================================

MDAnalysis reads coordinates from PDBQT_ files and additional optional
data such as B-factors, partial charge and AutoDock_ atom types.  It
is also possible to substitute a PDBQT file for a PSF file in order to
define the list of atoms (but no connectivity information will be
available in this case).

.. _PDBQT:
   https://autodock.scripps.edu/wp-content/uploads/sites/56/2021/10/AutoDock4.2.6_UserGuide.pdf
.. _AutoDock:
   http://autodock.scripps.edu/
"""

import os
import errno
import itertools
import numpy as np
import warnings

from ..lib import util
from . import base


class PDBQTReader(base.SingleFrameReaderBase):
    """PDBQTReader that reads a PDBQT-formatted file, no frills.

    Records read:
     - CRYST1 for unitcell A,B,C, alpha,beta,gamm
     - ATOM. HETATM for x,y,z

    Original `PDB format documentation`_ with `AutoDOCK extensions`_


    .. _PDB format documentation:
       http://www.wwpdb.org/documentation/file-format-content/format32/v3.2.html
    .. _AutoDOCK extensions:
       http://autodock.scripps.edu/faqs-help/faq/what-is-the-format-of-a-pdbqt-file

    =============  ============  ===========  =============================================
    COLUMNS        DATA  TYPE    FIELD        DEFINITION
    =============  ============  ===========  =============================================
    1 -  6         Record name   "CRYST1"
    7 - 15         Real(9.3)     a              a (Angstroms).
    16 - 24        Real(9.3)     b              b (Angstroms).
    25 - 33        Real(9.3)     c              c (Angstroms).
    34 - 40        Real(7.2)     alpha          alpha (degrees).
    41 - 47        Real(7.2)     beta           beta (degrees).
    48 - 54        Real(7.2)     gamma          gamma (degrees).

    1 -  6         Record name   "ATOM  "
    7 - 11         Integer       serial       Atom  serial number.
    13 - 16        Atom          name         Atom name.
    17             Character     altLoc       Alternate location indicator. IGNORED
    18 - 21        Residue name  resName      Residue name.
    22             Character     chainID      Chain identifier.
    23 - 26        Integer       resSeq       Residue sequence number.
    27             AChar         iCode        Code for insertion of residues. IGNORED
    31 - 38        Real(8.3)     x            Orthogonal coordinates for X in Angstroms.
    39 - 46        Real(8.3)     y            Orthogonal coordinates for Y in Angstroms.
    47 - 54        Real(8.3)     z            Orthogonal coordinates for Z in Angstroms.
    55 - 60        Real(6.2)     occupancy    Occupancy.
    61 - 66        Real(6.2)     tempFactor   Temperature  factor.
    67 - 70        LString(4)    footnote     Usually blank. IGNORED.
    71 - 76        Real(6.4)     partialChrg  Gasteiger PEOE partial charge *q*.
    79 - 80        LString(2)    atomType     AutoDOCK atom type *t*.
    =============  ============  ===========  =============================================

    We ignore torsion notation and just pull the partial charge and atom type columns::

         COMPND    NSC7810
         REMARK  3 active torsions:
         REMARK  status: ('A' for Active; 'I' for Inactive)
         REMARK    1  A    between atoms: A7_7  and  C22_23
         REMARK    2  A    between atoms: A9_9  and  A11_11
         REMARK    3  A    between atoms: A17_17  and  C21_21
         ROOT
         123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789. (column reference)
         ATOM      1  A1  INH I           1.054   3.021   1.101  0.00  0.00     0.002 A
         ATOM      2  A2  INH I           1.150   1.704   0.764  0.00  0.00     0.012 A
         ATOM      3  A3  INH I          -0.006   0.975   0.431  0.00  0.00    -0.024 A
         ATOM      4  A4  INH I           0.070  -0.385   0.081  0.00  0.00     0.012 A
         ATOM      5  A5  INH I          -1.062  -1.073  -0.238  0.00  0.00     0.002 A
         ATOM      6  A6  INH I          -2.306  -0.456  -0.226  0.00  0.00     0.019 A
         ATOM      7  A7  INH I          -2.426   0.885   0.114  0.00  0.00     0.052 A
         ATOM      8  A8  INH I          -1.265   1.621   0.449  0.00  0.00     0.002 A
         ATOM      9  A9  INH I          -1.339   2.986   0.801  0.00  0.00    -0.013 A
         ATOM     10  A10 INH I          -0.176   3.667   1.128  0.00  0.00     0.013 A
         ENDROOT
         BRANCH   9  11
         ATOM     11  A11 INH I          -2.644   3.682   0.827  0.00  0.00    -0.013 A
         ATOM     12  A16 INH I          -3.007   4.557  -0.220  0.00  0.00     0.002 A
         ATOM     13  A12 INH I          -3.522   3.485   1.882  0.00  0.00     0.013 A
         ATOM     14  A15 INH I          -4.262   5.209  -0.177  0.00  0.00    -0.024 A
         ATOM     15  A17 INH I          -2.144   4.784  -1.319  0.00  0.00     0.052 A
         ATOM     16  A14 INH I          -5.122   4.981   0.910  0.00  0.00     0.012 A
         ATOM     17  A20 INH I          -4.627   6.077  -1.222  0.00  0.00     0.012 A
         ATOM     18  A13 INH I          -4.749   4.135   1.912  0.00  0.00     0.002 A
         ATOM     19  A19 INH I          -3.777   6.285  -2.267  0.00  0.00     0.002 A
         ATOM     20  A18 INH I          -2.543   5.650  -2.328  0.00  0.00     0.019 A
         BRANCH  15  21
         ATOM     21  C21 INH I          -0.834   4.113  -1.388  0.00  0.00     0.210 C
         ATOM     22  O1  INH I          -0.774   2.915  -1.581  0.00  0.00    -0.644 OA
         ATOM     23  O3  INH I           0.298   4.828  -1.237  0.00  0.00    -0.644 OA
         ENDBRANCH  15  21
         ENDBRANCH   9  11
         BRANCH   7  24
         ATOM     24  C22 INH I          -3.749   1.535   0.125  0.00  0.00     0.210 C
         ATOM     25  O2  INH I          -4.019   2.378  -0.708  0.00  0.00    -0.644 OA
         ATOM     26  O4  INH I          -4.659   1.196   1.059  0.00  0.00    -0.644 OA
         ENDBRANCH   7  24
         TORSDOF 3
         123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789. (column reference)

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    """

    format = "PDBQT"
    units = {"time": None, "length": "Angstrom"}

    def _read_first_frame(self):
        coords = []
        unitcell = np.zeros(6, dtype=np.float32)
        with util.openany(self.filename) as pdbfile:
            for line in pdbfile:
                # Should only break at the 'END' of a model definition
                # and prevent premature exit for a torsion termination
                # , eg, ENDBRANCH
                if line.startswith("END\n"):
                    break
                if line.startswith("CRYST1"):
                    # lengths
                    x, y, z = np.float32(
                        (line[6:15], line[15:24], line[24:33])
                    )
                    # angles
                    A, B, G = np.float32(
                        (line[33:40], line[40:47], line[47:54])
                    )
                    unitcell[:] = x, y, z, A, B, G
                if line.startswith(("ATOM", "HETATM")):
                    # convert all entries at the end once for optimal speed
                    coords.append([line[30:38], line[38:46], line[46:54]])
        self.n_atoms = len(coords)
        self.ts = self._Timestep.from_coordinates(coords, **self._ts_kwargs)
        self.ts.dimensions = unitcell
        self.ts.frame = 0  # 0-based frame number
        if self.convert_units:
            # in-place !
            self.convert_pos_from_native(self.ts._pos)
            if self.ts.dimensions is not None:
                self.convert_pos_from_native(self.ts.dimensions[:3])

    def Writer(self, filename, **kwargs):
        """Returns a permissive (simple) PDBQTWriter for *filename*.

        Parameters
        ----------
        filename : str
            filename of the output PDBQT file

        Returns
        -------
        :class:`PDBQTWriter`

        """
        return PDBQTWriter(filename, **kwargs)


class PDBQTWriter(base.WriterBase):
    """PDBQT writer that implements a subset of the PDB_ 3.2 standard and the PDBQT_ spec.

    .. _PDB: http://www.wwpdb.org/documentation/file-format-content/format32/v3.2.html
    .. _PDBQT: https://autodock.scripps.edu/wp-content/uploads/sites/56/2021/10/AutoDock4.2.6_UserGuide.pdf

    .. versionchanged:: 2.6.0
       Files are now written in `wt` mode, and keep extensions, allowing
       for files to be written under compressed formats
    """

    fmt = {
        "ATOM": (
            "ATOM  {serial:5d} {name:<4.4s} {resName:<4.4s}"
            "{chainID:1.1s}{resSeq:4d}{iCode:1.1s}"
            "   {pos[0]:8.3f}{pos[1]:8.3f}{pos[2]:8.3f}{occupancy:6.2f}"
            "{tempFactor:6.2f}    {charge:< 1.3f} {element:<2.2s}\n"
        ),
        "REMARK": "REMARK     {0}\n",
        "TITLE": "TITLE     {0}\n",
        "CRYST1": (
            "CRYST1{box[0]:9.3f}{box[1]:9.3f}{box[2]:9.3f}"
            "{ang[0]:7.2f}{ang[1]:7.2f}{ang[2]:7.2f} "
            "{spacegroup:<11s}{zvalue:4d}\n"
        ),
    }
    format = "PDBQT"
    units = {"time": None, "length": "Angstrom"}
    pdb_coor_limits = {"min": -999.9995, "max": 9999.9995}

    def __init__(self, filename, **kwargs):
        self.filename = util.filename(filename, ext="pdbqt", keep=True)
        self.pdb = util.anyopen(self.filename, "wt")

    def close(self):
        self.pdb.close()

    def write(self, selection, frame=None):
        """Write selection at current trajectory frame to file.

        Parameters
        ----------
        selection : AtomGroup
            The selection to be written
        frame : int (optional)
            optionally move to frame index `frame` before writing; the default
            is to write the current trajectory frame

        Note
        ----
        The last letter of the
        :attr:`~MDAnalysis.core.groups.Atom.segid` is used as the PDB
        chainID.


        .. versionchanged:: 0.11.0
           Frames now 0-based instead of 1-based

        """
        try:
            u = selection.universe
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        if frame is not None:
            u.trajectory[frame]  # advance to frame
        else:
            try:
                frame = u.trajectory.ts.frame
            except AttributeError:
                frame = 0  # should catch cases when we are analyzing a single PDB (?)

        atoms = selection.atoms  # make sure to use atoms (Issue 46)
        coor = (
            atoms.positions
        )  # can write from selection == Universe (Issue 49)

        # Check attributes
        attrs = {}
        missing_topology = []
        for attr, dflt in (
            ("altLocs", " "),
            ("charges", 0.0),
            ("icodes", " "),
            ("names", "X"),
            ("occupancies", 1.0),
            ("resids", 1),
            ("resnames", "UNK"),
            ("tempfactors", 0.0),
            ("types", "  "),
        ):
            try:
                attrs[attr] = getattr(atoms, attr)
            except AttributeError:
                attrs[attr] = itertools.cycle((dflt,))
                missing_topology.append(attr)
        # Order of preference: chainids -> segids -> blank string
        try:
            attrs["chainids"] = atoms.chainids
        except AttributeError:
            try:
                attrs["chainids"] = atoms.segids
            except AttributeError:
                attrs["chainids"] = itertools.cycle((" ",))
                missing_topology.append("chainids")
        if missing_topology:
            warnings.warn(
                "Supplied AtomGroup was missing the following attributes: "
                "{miss}. These will be written with default values. "
                "".format(miss=", ".join(missing_topology))
            )

        # check if any coordinates are illegal (coordinates are already
        # in Angstroem per package default)
        if not self.has_valid_coordinates(self.pdb_coor_limits, coor):
            self.close()
            try:
                os.remove(self.filename)
            except OSError as err:
                if err.errno == errno.ENOENT:
                    pass
            raise ValueError(
                "PDB files must have coordinate values between {0:.3f}"
                " and {1:.3f} Angstroem: No file was written."
                "".format(
                    self.pdb_coor_limits["min"], self.pdb_coor_limits["max"]
                )
            )

        # Write title record
        # http://www.wwpdb.org/documentation/file-format-content/format32/sect2.html
        line = "FRAME " + str(frame) + " FROM " + str(u.trajectory.filename)
        self.pdb.write(self.fmt["TITLE"].format(line))

        # Write CRYST1 record
        # http://www.wwpdb.org/documentation/file-format-content/format32/sect8.html
        box = self.convert_dimensions_to_unitcell(u.trajectory.ts)
        self.pdb.write(
            self.fmt["CRYST1"].format(
                box=box[:3], ang=box[3:], spacegroup="P 1", zvalue=1
            )
        )

        # Write atom records
        # http://www.wwpdb.org/documentation/file-format-content/format32/sect9.html
        for serial, (
            pos,
            name,
            resname,
            chainid,
            resid,
            icode,
            occupancy,
            tempfactor,
            charge,
            element,
        ) in enumerate(
            zip(
                coor,
                attrs["names"],
                attrs["resnames"],
                attrs["chainids"],
                attrs["resids"],
                attrs["icodes"],
                attrs["occupancies"],
                attrs["tempfactors"],
                attrs["charges"],
                attrs["types"],
            ),
            start=1,
        ):
            serial = util.ltruncate_int(serial, 5)  # check for overflow here?
            resid = util.ltruncate_int(resid, 4)
            name = name[:4]
            if len(name) < 4:
                name = " " + name  # customary to start in column 14
            chainid = chainid.strip()[-1:]  # take the last character

            self.pdb.write(
                self.fmt["ATOM"].format(
                    serial=serial,
                    name=name,
                    resName=resname,
                    chainID=chainid,
                    resSeq=resid,
                    iCode=icode,
                    pos=pos,
                    occupancy=occupancy,
                    tempFactor=tempfactor,
                    charge=charge,
                    element=element,
                )
            )

        self.close()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
:mod:`MDAnalysis.lib` --- Access to lower level routines
================================================================
"""

__all__ = [
    "log",
    "transformations",
    "util",
    "mdamath",
    "distances",
    "NeighborSearch",
    "formats",
    "pkdtree",
    "nsgrid",
]

from . import log
from . import transformations
from . import util
from . import mdamath
from . import distances  # distances relies on mdamath
from . import NeighborSearch
from . import formats
from . import pkdtree
from . import nsgrid
from .picklable_file_io import (
    FileIOPicklable,
    BufferIOPicklable,
    TextIOPicklable,
)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""MOL2 file format --- :mod:`MDAnalysis.coordinates.MOL2`
========================================================

Classes to work with Tripos_ molecule structure format (MOL2_) coordinate and
topology files. Used, for instance, by the DOCK_ docking code.


Example for working with mol2 files
-----------------------------------

To open a mol2, remove all hydrogens and save as a new file, use the following::

  u = Universe("Molecule.mol2")
  gr = u.select_atoms("not name H*")
  print(len(u.atoms), len(gr))
  gr.write("Molecule_noh.mol2")

.. _MOL2: http://www.tripos.com/data/support/mol2.pdf
.. _Tripos: http://www.tripos.com/
.. _DOCK: http://dock.compbio.ucsf.edu/


See Also
--------

rdkit: rdkit_ is an open source cheminformatics toolkit with more mol2
       functionality


.. _rdkit: http://www.rdkit.org/docs/GettingStartedInPython.html


Classes
-------

.. autoclass:: MOL2Reader
   :members:

.. autoclass:: MOL2Writer
   :members:


Notes
-----

* The MDAnalysis :class:`MOL2Reader` and :class:`MOL2Writer` only handle the
  MOLECULE, SUBSTRUCTURE, ATOM, and BOND record types. Other records are not
  currently read or preserved on writing.
* As the CRYSIN record type is not parsed / written, MOL2 systems always have
  dimensions set to ``None`` and dimensionless MOL2 files are written.


MOL2 format notes
-----------------

* MOL2 Format Specification:  (http://www.tripos.com/data/support/mol2.pdf)
* Example file (http://www.tripos.com/mol2/mol2_format3.html)::

    #    Name: benzene
    #    Creating user name: tom
    #    Creation time: Wed Dec 28 00:18:30 1988

    #    Modifying user name: tom
    #    Modification time: Wed Dec 28 00:18:30 1988

    @<TRIPOS>MOLECULE
    benzene
    12 12 1  0   0
    SMALL
    NO_CHARGES


    @<TRIPOS>ATOM
    1   C1  1.207   2.091   0.000   C.ar    1   BENZENE 0.000
    2   C2  2.414   1.394   0.000   C.ar    1   BENZENE 0.000
    3   C3  2.414   0.000   0.000   C.ar    1   BENZENE 0.000
    4   C4  1.207   -0.697  0.000   C.ar    1   BENZENE 0.000
    5   C5  0.000   0.000   0.000   C.ar    1   BENZENE 0.000
    6   C6  0.000   1.394   0.000   C.ar    1   BENZENE 0.000
    7   H1  1.207   3.175   0.000   H   1   BENZENE 0.000
    8   H2  3.353   1.936   0.000   H   1   BENZENE 0.000
    9   H3  3.353   -0.542  0.000   H   1   BENZENE 0.000
    10  H4  1.207   -1.781  0.000   H   1   BENZENE 0.000
    11  H5  -0.939  -0.542  0.000   H   1   BENZENE 0.000
    12  H6  -0.939  1.936   0.000   H   1   BENZENE 0.000
    @<TRIPOS>BOND
    1   1   2   ar
    2   1   6   ar
    3   2   3   ar
    4   3   4   ar
    5   4   5   ar
    6   5   6   ar
    7   1   7   1
    8   2   8   1
    9   3   9   1
    10  4   10  1
    11  5   11  1
    12  6   12  1
   @<TRIPOS>SUBSTRUCTURE
    1   BENZENE 1   PERM    0   ****    ****    0   ROOT

"""
import numpy as np

from . import base
from ..lib import util
from MDAnalysis.lib.util import store_init_arguments


class MOL2Reader(base.ReaderBase):
    """Reader for MOL2 structure format.

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based.
       MOL2 now reuses the same Timestep object for every frame,
       previously created a new instance of Timestep each frame.
    .. versionchanged:: 0.20.0
       Allows for comments at top of file.
       Ignores status bit strings
    .. versionchanged:: 2.0.0
       Bonds attribute is not added if no bonds are present in MOL2 file
    .. versionchanged:: 2.2.0
       Read MOL2 files with optional columns omitted.
    """

    format = "MOL2"
    units = {"time": None, "length": "Angstrom"}

    @store_init_arguments
    def __init__(self, filename, **kwargs):
        """Read coordinates from `filename`.


        Parameters
        ----------
        filename : str or NamedStream
             name of the mol2 file or stream
        """
        super(MOL2Reader, self).__init__(filename, **kwargs)

        self.n_atoms = None

        blocks = []

        with util.openany(filename) as f:
            for i, line in enumerate(f):
                # found new molecules
                if "@<TRIPOS>MOLECULE" in line:
                    blocks.append({"start_line": i, "lines": []})
                if len(blocks):
                    blocks[-1]["lines"].append(line)
        self.n_frames = len(blocks)
        self.frames = blocks

        sections, coords = self.parse_block(blocks[0])
        self.n_atoms = len(coords)

        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        self.ts = self._read_frame(0)

    def parse_block(self, block):
        sections = {}
        cursor = None
        for line in block["lines"]:
            if "@<TRIPOS>" in line:
                cursor = line.split("@<TRIPOS>")[1].strip().lower()
                sections[cursor] = []
                continue
            elif line.startswith("#") or line == "\n":
                continue
            sections[cursor].append(line)

        atom_lines = sections["atom"]
        if not len(atom_lines):
            raise Exception(
                "The mol2 (starting at line {0}) block has no atoms"
                "".format(block["start_line"])
            )
        elif self.n_atoms is None:
            # First time round, remember the number of atoms
            self.n_atoms = len(atom_lines)
        elif len(atom_lines) != self.n_atoms:
            raise ValueError(
                "MOL2Reader assumes that the number of atoms remains unchanged"
                " between frames; the current "
                "frame has {0}, the next frame has {1} atoms"
                "".format(self.n_atoms, len(atom_lines))
            )

        coords = np.zeros((self.n_atoms, 3), dtype=np.float32)
        for i, a in enumerate(atom_lines):
            aid, name, x, y, z, atom_type = a.split()[:6]

            # x, y, z = float(x), float(y), float(z)
            coords[i, :] = x, y, z

        return sections, coords

    def _read_next_timestep(self, ts=None):
        if ts is None:
            ts = self.ts
        else:
            # TODO: cleanup _read_frame() to use a "free" Timestep
            raise NotImplementedError("MOL2Reader cannot assign to a timestep")
        frame = self.frame + 1
        return self._read_frame(frame)

    def _read_frame(self, frame):
        try:
            block = self.frames[frame]
        except IndexError:
            errmsg = (
                f"Invalid frame {frame} for trajectory with length "
                f"{len(self)}"
            )
            raise IOError(errmsg) from None

        sections, coords = self.parse_block(block)

        for sect in ["molecule", "substructure"]:
            try:
                self.ts.data[sect] = sections[sect]
            except KeyError:
                pass

        self.ts.positions = np.array(coords, dtype=np.float32)

        if self.convert_units:
            # in-place !
            self.convert_pos_from_native(self.ts._pos)
        self.ts.frame = frame

        return self.ts

    def _reopen(self):
        # Make frame think it's before start, so calling next
        # reads first frame
        self.ts.frame = -1


class MOL2Writer(base.WriterBase):
    """mol2 format writer

    Write a file in the Tripos_ molecule structure format (MOL2_).

    Note
    ----
    :class:`MOL2Writer` can only be used to write out previously loaded MOL2 files.
    If you're trying to convert, for example, a PDB file to MOL you should
    use other tools, like rdkit_.

    Here is an example how to use rdkit_ to convert a PDB to MOL::

      from rdkit import Chem
      mol = Chem.MolFromPDBFile("molecule.pdb", removeHs=False)
      Chem.MolToMolFile(mol, "molecule.mol" )

    MOL2 writer is currently not available for rdkit master. It requires SYBYL
    atomtype generation.
    See page 7 for list of SYBYL atomtypes
    (http://tripos.com/tripos_resources/fileroot/pdfs/mol2_format2.pdf).


    .. _rdkit: http://www.rdkit.org/docs/GettingStartedInPython.html

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based

    """

    format = "MOL2"
    multiframe = True
    units = {"time": None, "length": "Angstrom"}

    def __init__(self, filename, n_atoms=None, convert_units=True):
        """Create a new MOL2Writer

        Parameters
        ----------
        filename: str
            name of output file
        convert_units: bool (optional)
            units are converted to the MDAnalysis base format; [``True``]
        """
        self.filename = filename
        self.convert_units = (
            convert_units  # convert length and time to base units
        )

        self.frames_written = 0

        self.file = util.anyopen(self.filename, "w")  # open file on init

    def close(self):
        self.file.close()

    def encode_block(self, obj):
        """
        Parameters
        ----------
        obj : AtomGroup or Universe
        """
        # Issue 2717
        try:
            obj = obj.atoms
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        traj = obj.universe.trajectory
        ts = traj.ts

        try:
            molecule = ts.data["molecule"]
        except KeyError:
            errmsg = "MOL2Writer cannot currently write non MOL2 data"
            raise NotImplementedError(errmsg) from None

        # Need to remap atom indices to 1 based in this selection
        mapping = {a: i for i, a in enumerate(obj.atoms, start=1)}

        # only write bonds if the Bonds attribute exists (Issue #3057)
        if hasattr(obj, "bonds"):
            # Grab only bonds between atoms in the obj
            # ie none that extend out of it
            bondgroup = obj.intra_bonds
            bonds = sorted((b[0], b[1], b.order) for b in bondgroup)
            bond_lines = ["@<TRIPOS>BOND"]
            bls = [
                "{0:>5} {1:>5} {2:>5} {3:>2}".format(
                    bid, mapping[atom1], mapping[atom2], order
                )
                for bid, (atom1, atom2, order) in enumerate(bonds, start=1)
            ]
            bond_lines.extend(bls)
            bond_lines.append("\n")
            bond_lines = "\n".join(bond_lines)
        else:
            bondgroup = []
            bond_lines = ""

        atom_lines = ["@<TRIPOS>ATOM"]
        atom_lines.extend(
            "{0:>4} {1:>4} {2:>13.4f} {3:>9.4f} {4:>9.4f}"
            " {5:>4} {6} {7} {8:>7.4f}"
            "".format(
                mapping[a],
                a.name,
                a.position[0],
                a.position[1],
                a.position[2],
                a.type,
                a.resid,
                a.resname,
                a.charge,
            )
            for a in obj.atoms
        )
        atom_lines.append("\n")
        atom_lines = "\n".join(atom_lines)

        try:
            substructure = ["@<TRIPOS>SUBSTRUCTURE\n"] + ts.data[
                "substructure"
            ]
        except KeyError:
            substructure = ""

        check_sums = molecule[1].split()
        check_sums[0], check_sums[1] = str(len(obj.atoms)), str(len(bondgroup))

        # prevent behavior change between repeated calls
        # see gh-2678
        molecule_0_store = molecule[0]
        molecule_1_store = molecule[1]

        molecule[1] = "{0}\n".format(" ".join(check_sums))
        molecule.insert(0, "@<TRIPOS>MOLECULE\n")

        return_val = (
            "".join(molecule) + atom_lines + bond_lines + "".join(substructure)
        )

        molecule[0] = molecule_0_store
        molecule[1] = molecule_1_store
        return return_val

    def _write_next_frame(self, obj):
        """Write a new frame to the MOL2 file.

        Parameters
        ----------
        obj : AtomGroup or Universe


        .. versionchanged:: 1.0.0
            Renamed from `write_next_timestep` to `_write_next_frame`.
        """
        block = self.encode_block(obj)
        self.file.writelines(block)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import pytest
import numpy as np

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import TXYZ, ARC, ARC_PBC
from numpy.testing import assert_equal, assert_allclose


class TestTXYZParser(ParserBase):
    parser = mda.topology.TXYZParser.TXYZParser
    guessed_attrs = ["masses"]
    expected_attrs = ["ids", "names", "bonds", "types", "elements"]

    expected_n_residues = 1
    expected_n_atoms = 9
    expected_n_segments = 1

    def test_number_of_bonds(self, top):
        assert len(top.bonds.values) == 8

    @pytest.fixture(params=(TXYZ, ARC))
    def filename(self, request):
        return request.param

    def test_atom_type_type(self, top):
        """
        ``u.atoms.types`` contains strings.

        See `issue #2435 <https://github.com/MDAnalysis/mdanalysis/issues/2435>
        """
        types = top.types.values
        type_is_str = [isinstance(atom_type, str) for atom_type in types]
        assert all(type_is_str)


def test_TXYZ_elements():
    """The test checks whether elements attribute are assigned
    properly given a TXYZ file with valid elements record.
    """
    u = mda.Universe(TXYZ, format="TXYZ")
    element_list = np.array(
        ["C", "H", "H", "O", "H", "C", "H", "H", "H"], dtype=object
    )
    assert_equal(u.atoms.elements, element_list)


def test_missing_elements_noattribute():
    """Check that:

    1) a warning is raised if elements are missing
    2) the elements attribute is not set
    """
    wmsg = (
        "Element information is missing, elements attribute "
        "will not be populated. If needed these can be "
    )
    with pytest.warns(UserWarning, match=wmsg):
        u = mda.Universe(ARC_PBC)
    with pytest.raises(AttributeError):
        _ = u.atoms.elements


def test_guessed_masses():
    u = mda.Universe(TXYZ)
    expected = [
        12.011,
        1.008,
        1.008,
        15.999,
        1.008,
        12.011,
        1.008,
        1.008,
        1.008,
    ]
    assert_allclose(u.atoms.masses, expected)



#!/usr/bin/env python
# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Setuptools-based setup script for MDAnalysis.

A working installation of NumPy <http://numpy.scipy.org> is required.

For a basic installation just type the command::

  pip install .

For more in-depth instructions, see the installation section at the
MDAnalysis User Guide:

  https://userguide.mdanalysis.org/stable/installation.html

Also free to ask on GitHub Discussions for help:

  https://github.com/MDAnalysis/mdanalysis/discussions/categories/installation

"""

from setuptools import setup, Extension, find_packages
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler
from packaging.version import Version
import codecs
import os
import sys
import re
import shutil
import tempfile
import warnings
import platform

import configparser
from subprocess import getoutput

# NOTE: keep in sync with MDAnalysis.__version__ in version.py
RELEASE = "2.10.0"

is_release = "dev" not in RELEASE

# Handle cython modules
try:
    # cython has to be >=0.16 <0.28 to support cython.parallel
    # minimum cython version now set to 0.28 to match pyproject.toml
    import Cython
    from Cython.Build import cythonize

    cython_found = True

    required_version = "0.28"
    if not Version(Cython.__version__) >= Version(required_version):
        # We don't necessarily die here. Maybe we already have
        #  the cythonized '.c' files.
        print(
            "Cython version {0} was found but won't be used: version {1} "
            "or greater is required because it offers a handy "
            "parallelization module".format(
                Cython.__version__, required_version
            )
        )
        cython_found = False
    cython_linetrace = bool(os.environ.get("CYTHON_TRACE_NOGIL", False))
except ImportError:
    cython_found = False
    if not is_release:
        print("*** package: Cython not found ***")
        print("MDAnalysis requires cython for development builds")
        sys.exit(1)
    cython_linetrace = False


def abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


class Config(object):
    """Config wrapper class to get build options

    This class looks for options in the environment variables and the
    'setup.cfg' file. The order how we look for an option is.

    1. Environment Variable
    2. set in 'setup.cfg'
    3. given default

    Environment variables should start with 'MDA_' and be all uppercase.
    Values passed to environment variables are checked (case-insensitively)
    for specific strings with boolean meaning: 'True' or '1' will cause `True`
    to be returned. '0' or 'False' cause `False` to be returned.

    """

    def __init__(self, fname="setup.cfg"):
        fname = abspath(fname)
        if os.path.exists(fname):
            self.config = configparser.ConfigParser()
            self.config.read(fname)

    def get(self, option_name, default=None):
        environ_name = "MDA_" + option_name.upper()
        if environ_name in os.environ:
            val = os.environ[environ_name]
            if val.upper() in ("1", "TRUE"):
                return True
            elif val.upper() in ("0", "FALSE"):
                return False
            return val
        try:
            option = self.config.get("options", option_name)
            return option
        except configparser.NoOptionError:
            return default


class MDAExtension(Extension, object):
    """Derived class to cleanly handle setup-time (numpy) dependencies."""

    # The only setup-time numpy dependency comes when setting up its
    #  include dir.
    # The actual numpy import and call can be delayed until after pip
    #  has figured it must install numpy.
    # This is accomplished by passing the get_numpy_include function
    #  as one of the include_dirs. This derived Extension class takes
    #  care of calling it when needed.
    def __init__(self, name, sources, *args, **kwargs):
        self._mda_include_dirs = []
        # don't abspath sources else packaging fails on Windows (issue #3129)
        super(MDAExtension, self).__init__(name, sources, *args, **kwargs)

    @property
    def include_dirs(self):
        if not self._mda_include_dirs:
            for item in self._mda_include_dir_args:
                try:
                    self._mda_include_dirs.append(item())  # The numpy callable
                except TypeError:
                    item = abspath(item)
                    self._mda_include_dirs.append((item))
        return self._mda_include_dirs

    @include_dirs.setter
    def include_dirs(self, val):
        self._mda_include_dir_args = val


def get_numpy_include():
    # Obtain the numpy include directory. This logic works across numpy
    # versions.
    # setuptools forgets to unset numpy's setup flag and we get a crippled
    # version of it unless we do it ourselves.
    import builtins

    builtins.__NUMPY_SETUP__ = False
    try:
        import numpy as np
    except ImportError:
        print('*** package "numpy" not found ***')
        print(
            "MDAnalysis requires a version of NumPy (>=1.21.0), even for setup."
        )
        print(
            "Please get it from http://numpy.scipy.org/ or install it through "
            "your package manager."
        )
        sys.exit(-1)
    return np.get_include()


def hasfunction(cc, funcname, include=None, extra_postargs=None):
    # From http://stackoverflow.com/questions/
    #            7018879/disabling-output-when-compiling-with-distutils
    tmpdir = tempfile.mkdtemp(prefix="hasfunction-")
    devnull = oldstderr = None
    try:
        try:
            fname = os.path.join(tmpdir, "funcname.c")
            with open(fname, "w") as f:
                if include is not None:
                    f.write("#include {0!s}\n".format(include))
                f.write("int main(void) {\n")
                f.write("    {0!s};\n".format(funcname))
                f.write("}\n")
            # Redirect stderr to /dev/null to hide any error messages
            # from the compiler.
            # This will have to be changed if we ever have to check
            # for a function on Windows.
            devnull = open("/dev/null", "w")
            oldstderr = os.dup(sys.stderr.fileno())
            os.dup2(devnull.fileno(), sys.stderr.fileno())
            objects = cc.compile(
                [fname], output_dir=tmpdir, extra_postargs=extra_postargs
            )
            cc.link_executable(objects, os.path.join(tmpdir, "a.out"))
        except Exception:
            return False
        return True
    finally:
        if oldstderr is not None:
            os.dup2(oldstderr, sys.stderr.fileno())
        if devnull is not None:
            devnull.close()
        shutil.rmtree(tmpdir)


def detect_openmp():
    """Does this compiler support OpenMP parallelization?"""
    print("Attempting to autodetect OpenMP support... ", end="")
    compiler = new_compiler()
    customize_compiler(compiler)
    compiler.add_library("gomp")
    include = "<omp.h>"
    extra_postargs = ["-fopenmp"]
    hasopenmp = hasfunction(
        compiler,
        "omp_get_num_threads()",
        include=include,
        extra_postargs=extra_postargs,
    )
    if hasopenmp:
        print("Compiler supports OpenMP")
    else:
        print("Did not detect OpenMP support.")
    return hasopenmp


def using_clang():
    """Will we be using a clang compiler?"""
    compiler = new_compiler()
    customize_compiler(compiler)
    compiler_ver = getoutput("{0} -v".format(compiler.compiler[0]))
    if "Spack GCC" in compiler_ver:
        # when gcc toolchain is built from source with spack
        # using clang, the 'clang' string may be present in
        # the compiler metadata, but it is not clang
        is_clang = False
    elif "clang" in compiler_ver:
        # by default, Apple will typically alias gcc to
        # clang, with some mention of 'clang' in the
        # metadata
        is_clang = True
    else:
        is_clang = False
    return is_clang


def extensions(config):
    # usually (except coming from release tarball) cython files must be generated
    use_cython = config.get("use_cython", default=cython_found)
    use_openmp = config.get("use_openmp", default=True)
    annotate_cython = config.get("annotate_cython", default=False)

    extra_compile_args = [
        "-std=c11",
        "-O3",
        "-funroll-loops",
        "-fsigned-zeros",
    ]  # see #2722
    define_macros = []
    if config.get("debug_cflags", default=False):
        extra_compile_args.extend(["-Wall", "-pedantic"])
        define_macros.extend([("DEBUG", "1")])

    # encore is sensitive to floating point accuracy, especially on non-x86
    # to avoid reducing optimisations on everything, we make a set of compile
    # args specific to encore see #2997 for an example of this.
    encore_compile_args = [a for a in extra_compile_args if "O3" not in a]
    if platform.machine() == "aarch64" or platform.machine() == "ppc64le":
        encore_compile_args.append("-O1")
    else:
        encore_compile_args.append("-O3")

    # allow using custom c/c++ flags and architecture specific instructions.
    # This allows people to build optimized versions of MDAnalysis.
    # Do here so not included in encore
    extra_cflags = config.get("extra_cflags", default=False)
    if extra_cflags:
        flags = extra_cflags.split()
        extra_compile_args.extend(flags)

    cpp_extra_compile_args = [a for a in extra_compile_args if "std" not in a]
    cpp_extra_compile_args.append("-std=c++11")
    cpp_extra_link_args = []
    # needed to specify c++ runtime library on OSX
    if platform.system() == "Darwin" and using_clang():
        cpp_extra_compile_args.append("-stdlib=libc++")
        cpp_extra_compile_args.append("-mmacosx-version-min=10.9")
        cpp_extra_link_args.append("-stdlib=libc++")
        cpp_extra_link_args.append("-mmacosx-version-min=10.9")

    # Needed for large-file seeking under 32bit systems (for xtc/trr indexing
    # and access).
    largefile_macros = [
        ("_LARGEFILE_SOURCE", None),
        ("_LARGEFILE64_SOURCE", None),
        ("_FILE_OFFSET_BITS", "64"),
    ]

    has_openmp = detect_openmp()

    if use_openmp and not has_openmp:
        print("No openmp compatible compiler found default to serial build.")

    parallel_args = ["-fopenmp"] if has_openmp and use_openmp else []
    parallel_libraries = ["gomp"] if has_openmp and use_openmp else []
    parallel_macros = [("PARALLEL", None)] if has_openmp and use_openmp else []

    if use_cython:
        print("Will attempt to use Cython.")
        if not cython_found:
            print(
                "Couldn't find a Cython installation. "
                "Not recompiling cython extensions."
            )
            use_cython = False
    else:
        print("Will not attempt to use Cython.")

    source_suffix = ".pyx" if use_cython else ".c"
    cpp_source_suffix = ".pyx" if use_cython else ".cpp"

    # The callable is passed so that it is only evaluated at install time.

    include_dirs = [get_numpy_include]
    # Windows automatically handles math library linking
    # and will not build MDAnalysis if we try to specify one
    if os.name == "nt":
        mathlib = []
    else:
        mathlib = ["m"]

    if cython_linetrace:
        extra_compile_args.append("-DCYTHON_TRACE_NOGIL")
        cpp_extra_compile_args.append("-DCYTHON_TRACE_NOGIL")

    libdcd = MDAExtension(
        "MDAnalysis.lib.formats.libdcd",
        ["MDAnalysis/lib/formats/libdcd" + source_suffix],
        include_dirs=include_dirs + ["MDAnalysis/lib/formats/include"],
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
    )
    distances = MDAExtension(
        "MDAnalysis.lib.c_distances",
        ["MDAnalysis/lib/c_distances" + source_suffix],
        include_dirs=include_dirs + ["MDAnalysis/lib/include"],
        libraries=mathlib,
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
    )
    distances_omp = MDAExtension(
        "MDAnalysis.lib.c_distances_openmp",
        ["MDAnalysis/lib/c_distances_openmp" + source_suffix],
        include_dirs=include_dirs + ["MDAnalysis/lib/include"],
        libraries=mathlib + parallel_libraries,
        define_macros=define_macros + parallel_macros,
        extra_compile_args=parallel_args + extra_compile_args,
        extra_link_args=parallel_args,
    )
    qcprot = MDAExtension(
        "MDAnalysis.lib.qcprot",
        ["MDAnalysis/lib/qcprot" + source_suffix],
        include_dirs=include_dirs,
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
    )
    transformation = MDAExtension(
        "MDAnalysis.lib._transformations",
        ["MDAnalysis/lib/src/transformations/transformations.c"],
        libraries=mathlib,
        define_macros=define_macros,
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
    )
    libmdaxdr = MDAExtension(
        "MDAnalysis.lib.formats.libmdaxdr",
        sources=[
            "MDAnalysis/lib/formats/libmdaxdr" + source_suffix,
            "MDAnalysis/lib/formats/src/xdrfile.c",
            "MDAnalysis/lib/formats/src/xdrfile_xtc.c",
            "MDAnalysis/lib/formats/src/xdrfile_trr.c",
            "MDAnalysis/lib/formats/src/trr_seek.c",
            "MDAnalysis/lib/formats/src/xtc_seek.c",
        ],
        include_dirs=include_dirs
        + ["MDAnalysis/lib/formats/include", "MDAnalysis/lib/formats"],
        define_macros=largefile_macros + define_macros,
        extra_compile_args=extra_compile_args,
    )
    util = MDAExtension(
        "MDAnalysis.lib.formats.cython_util",
        sources=["MDAnalysis/lib/formats/cython_util" + source_suffix],
        include_dirs=include_dirs,
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
    )
    cutil = MDAExtension(
        "MDAnalysis.lib._cutil",
        sources=["MDAnalysis/lib/_cutil" + cpp_source_suffix],
        language="c++",
        libraries=mathlib,
        include_dirs=include_dirs + ["MDAnalysis/lib/include"],
        define_macros=define_macros,
        extra_compile_args=cpp_extra_compile_args,
        extra_link_args=cpp_extra_link_args,
    )
    augment = MDAExtension(
        "MDAnalysis.lib._augment",
        sources=["MDAnalysis/lib/_augment" + cpp_source_suffix],
        language="c++",
        include_dirs=include_dirs,
        define_macros=define_macros,
        extra_compile_args=cpp_extra_compile_args,
        extra_link_args=cpp_extra_link_args,
    )
    timestep = MDAExtension(
        "MDAnalysis.coordinates.timestep",
        sources=["MDAnalysis/coordinates/timestep" + cpp_source_suffix],
        language="c++",
        include_dirs=include_dirs,
        define_macros=define_macros,
        extra_compile_args=cpp_extra_compile_args,
        extra_link_args=cpp_extra_link_args,
    )

    encore_utils = MDAExtension(
        "MDAnalysis.analysis.encore.cutils",
        sources=["MDAnalysis/analysis/encore/cutils" + source_suffix],
        include_dirs=include_dirs,
        define_macros=define_macros,
        extra_compile_args=encore_compile_args,
    )
    ap_clustering = MDAExtension(
        "MDAnalysis.analysis.encore.clustering.affinityprop",
        sources=[
            "MDAnalysis/analysis/encore/clustering/affinityprop"
            + source_suffix,
            "MDAnalysis/analysis/encore/clustering/src/ap.c",
        ],
        include_dirs=include_dirs
        + ["MDAnalysis/analysis/encore/clustering/include"],
        libraries=mathlib,
        define_macros=define_macros,
        extra_compile_args=encore_compile_args,
    )
    spe_dimred = MDAExtension(
        "MDAnalysis.analysis.encore.dimensionality_reduction.stochasticproxembed",
        sources=[
            "MDAnalysis/analysis/encore/dimensionality_reduction/stochasticproxembed"
            + source_suffix,
            "MDAnalysis/analysis/encore/dimensionality_reduction/src/spe.c",
        ],
        include_dirs=include_dirs
        + ["MDAnalysis/analysis/encore/dimensionality_reduction/include"],
        libraries=mathlib,
        define_macros=define_macros,
        extra_compile_args=encore_compile_args,
    )
    nsgrid = MDAExtension(
        "MDAnalysis.lib.nsgrid",
        ["MDAnalysis/lib/nsgrid" + cpp_source_suffix],
        include_dirs=include_dirs + ["MDAnalysis/lib/include"],
        language="c++",
        define_macros=define_macros,
        extra_compile_args=cpp_extra_compile_args,
        extra_link_args=cpp_extra_link_args,
    )
    pre_exts = [
        libdcd,
        distances,
        distances_omp,
        qcprot,
        transformation,
        libmdaxdr,
        util,
        encore_utils,
        ap_clustering,
        spe_dimred,
        cutil,
        augment,
        nsgrid,
        timestep,
    ]

    cython_generated = []
    if use_cython:
        extensions = cythonize(
            pre_exts,
            annotate=annotate_cython,
            compiler_directives={
                "linetrace": cython_linetrace,
                "embedsignature": False,
                "language_level": "3",
            },
        )
        if cython_linetrace:
            print("Cython coverage will be enabled")
        for pre_ext, post_ext in zip(pre_exts, extensions):
            for source in post_ext.sources:
                if source not in pre_ext.sources:
                    cython_generated.append(source)
    else:
        # Let's check early for missing .c files
        extensions = pre_exts
        for ext in extensions:
            for source in ext.sources:
                if not (os.path.isfile(source) and os.access(source, os.R_OK)):
                    raise IOError(
                        "Source file '{}' not found. This might be "
                        "caused by a missing Cython install, or a "
                        "failed/disabled Cython build.".format(source)
                    )
    return extensions, cython_generated


def dynamic_author_list():
    """Generate __authors__ from AUTHORS

    This function generates authors.py that contains the list of the
    authors from the AUTHORS file. This avoids having that list maintained in
    several places. Note that AUTHORS is sorted chronologically while we want
    __authors__ in authors.py to be sorted alphabetically.

    The authors are written in AUTHORS as bullet points under the
    "Chronological list of authors" title.
    """
    authors = []
    with codecs.open(abspath("AUTHORS"), encoding="utf-8") as infile:
        # An author is a bullet point under the title "Chronological list of
        # authors". We first want move the cursor down to the title of
        # interest.
        for line_no, line in enumerate(infile, start=1):
            if line.rstrip() == "Chronological list of authors":
                break
        else:
            # If we did not break, it means we did not find the authors.
            raise IOError("EOF before the list of authors")
        # Skip the next line as it is the title underlining
        line = next(infile)
        line_no += 1
        if line[:4] != "----":
            raise IOError(
                "Unexpected content on line {0}, "
                'should be a string of "-".'.format(line_no)
            )
        # Add each bullet point as an author until the next title underlining
        for line in infile:
            if line[:4] in ("----", "====", "~~~~"):
                # The previous line was a title, hopefully it did not start as
                # a bullet point so it got ignored. Since we hit a title, we
                # are done reading the list of authors.
                break
            elif line.strip()[:2] == "- ":
                # This is a bullet point, so it should be an author name.
                name = line.strip()[2:].strip()
                authors.append(name)

    # So far, the list of authors is sorted chronologically. We want it
    # sorted alphabetically of the last name.
    authors.sort(key=lambda name: name.split()[-1])
    # Move Naveen and Elizabeth first, and Oliver last.
    authors.remove("Naveen Michaud-Agrawal")
    authors.remove("Elizabeth J. Denning")
    authors.remove("Oliver Beckstein")
    authors = (
        ["Naveen Michaud-Agrawal", "Elizabeth J. Denning"]
        + authors
        + ["Oliver Beckstein"]
    )

    # Write the authors.py file.
    out_path = abspath("MDAnalysis/authors.py")
    with codecs.open(out_path, "w", encoding="utf-8") as outfile:
        # Write the header
        header = """\
#-*- coding:utf-8 -*-

# This file is generated from the AUTHORS file during the installation process.
# Do not edit it as your changes will be overwritten.
"""
        print(header, file=outfile)

        # Write the list of authors as a python list
        template = "__authors__ = [\n{}\n]"
        author_string = ",\n".join(
            '    u"{}"'.format(name) for name in authors
        )
        print(template.format(author_string), file=outfile)


def long_description(readme):
    """Create reST SUMMARY file for PyPi."""

    with open(abspath(readme)) as summary:
        buffer = summary.read()
    # remove top heading that messes up pypi display
    m = re.search(
        "====*\n[^\n]*README[^\n]*\n=====*\n", buffer, flags=re.DOTALL
    )
    assert m, "README.rst does not contain a level-1 heading"
    return buffer[m.end() :]


if __name__ == "__main__":
    try:
        dynamic_author_list()
    except (OSError, IOError):
        warnings.warn("Cannot write the list of authors.")

    try:
        # when building from repository for creating the distribution
        LONG_DESCRIPTION = long_description("../README.rst")
    except OSError:
        # when building from a tar file for installation
        # (LONG_DESCRIPTION is not really needed)
        LONG_DESCRIPTION = "MDAnalysis -- https://www.mdanalysis.org/"

    config = Config()
    exts, cythonfiles = extensions(config)

    setup(
        name="MDAnalysis",
        version=RELEASE,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/x-rst",
        # currently unused & may become obsolte see setuptools #1569
        provides=["MDAnalysis"],
        ext_modules=exts,
        test_suite="MDAnalysisTests",
        tests_require=[
            "MDAnalysisTests=={0!s}".format(RELEASE),  # same as this release!
        ],
    )

    # Releases keep their cythonized stuff for shipping.
    if (
        not config.get("keep_cythonized", default=is_release)
        and not cython_linetrace
    ):
        for cythonized in cythonfiles:
            try:
                os.unlink(cythonized)
            except OSError as err:
                print(
                    "Warning: failed to delete cythonized file {0}: {1}. "
                    "Moving on.".format(cythonized, err.strerror)
                )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""LAMMPS DCD trajectory and DATA I/O  --- :mod:`MDAnalysis.coordinates.LAMMPS`
===============================================================================

Classes to read and write LAMMPS_ DCD binary trajectories, LAMMPS DATA files
and LAMMPS dump files.  Trajectories can be read regardless of
system-endianness as this is auto-detected.

LAMMPS can `write DCD`_ trajectories but unlike a `CHARMM trajectory`_
(which is often called a DCD even though CHARMM itself calls them
"trj") the time unit is not fixed to be the AKMA_ time unit (20 AKMA
is 0.978 picoseconds or 1 AKMA = 4.888821e-14 s) but can depend on
settings in LAMMPS. The most common case for biomolecular simulations
appears to be that the time step is recorded in femtoseconds (command
`units real`_ in the input file) and lengths in ångströms. Other cases
are unit-less Lennard-Jones time units.

This presents a problem for MDAnalysis because it cannot autodetect
the unit from the file. By default we are assuming that the unit for
length is the ångström and for the time is the femtosecond. If this is
not true then the user *should supply the appropriate units* in the
keywords *timeunit* and/or *lengthunit* to :class:`DCDWriter` and
:class:`~MDAnalysis.core.universe.Universe` (which then calls
:class:`DCDReader`).

Data file formats
-----------------

By default either the `atomic` or `full` atom styles are expected,
however this can be customised, see :ref:`atom_style_kwarg`.

Dump files
----------

The DumpReader expects ascii dump files written with the default
`LAMMPS dump format`_ of 'atom'


Example: Loading a LAMMPS simulation
------------------------------------

To load a LAMMPS simulation from a LAMMPS data file (using the
:class:`~MDAnalysis.topology.LAMMPSParser.DATAParser`) together with a
LAMMPS DCD with "*real*" provide the keyword *format="LAMMPS*"::

    >>> import MDAnalysis
    >>> from MDAnalysis.tests.datafiles import LAMMPSdata2, LAMMPSdcd2
    >>> u = MDAnalysis.Universe(LAMMPSdata2, LAMMPSdcd2, format="LAMMPS")

If the trajectory uses *units nano* then use

    >>> import MDAnalysis
    >>> from MDAnalysis.tests.datafiles import LAMMPSdata2, LAMMPSdcd2
    >>> u = MDAnalysis.Universe(LAMMPSdata2, LAMMPSdcd2, format="LAMMPS",
    ...                          lengthunit="nm", timeunit="ns")

To scan through a trajectory to find a desirable frame and write to a LAMMPS
data file,

   >>> import MDAnalysis
   >>> from MDAnalysis.tests.datafiles import LAMMPSdata2, LAMMPSdcd2
   >>> u = MDAnalysis.Universe(LAMMPSdata2, LAMMPSdcd2, format="LAMMPS",
   ...                          lengthunit="nm", timeunit="ns")
   >>> take_this_frame = False
   >>> for ts in u.trajectory:
   ...     # analyze frame
   ...     if ts.frame == 4:
   ...         take_this_frame = True
   ...     if take_this_frame == True:
   ...         with MDAnalysis.Writer('frame.data') as W:
   ...             W.write(u.atoms)
   ...         break

Note
----
Lennard-Jones units are not implemented. See :mod:`MDAnalysis.units`
for other recognized values and the documentation for the LAMMPS
`units command`_.

See Also
--------

   For further discussion follow the reports for `Issue 84`_ and `Issue 64`_.

.. _LAMMPS: https://www.lammps.org/
.. _write DCD: https://docs.lammps.org/dump.html
.. _CHARMM trajectory:
   http://www.charmm.org/documentation/c36b1/dynamc.html#%20Trajectory
.. _AKMA: http://www.charmm.org/documentation/c36b1/usage.html#%20AKMA
.. _units real: https://docs.lammps.org/units.html
.. _units command: https://docs.lammps.org/units.html
.. _`Issue 64`: https://github.com/MDAnalysis/mdanalysis/issues/64
.. _`Issue 84`: https://github.com/MDAnalysis/mdanalysis/issues/84
.. _`LAMMPS dump format`: https://docs.lammps.org/dump.html

Classes
-------

.. autoclass:: DCDReader
   :members:
   :inherited-members:
.. autoclass:: DCDWriter
   :members:
   :inherited-members:
.. autoclass:: DATAReader
   :members:
   :inherited-members:
.. autoclass:: DATAWriter
   :members:
   :inherited-members:
.. autoclass:: DumpReader
   :members:
   :inherited-members:

"""
import os
import numpy as np

from ..core.groups import requires
from ..lib import util, mdamath, distances
from ..lib.util import cached, store_init_arguments
from . import DCD
from .. import units
from ..topology.LAMMPSParser import DATAParser
from ..exceptions import NoDataError
from . import base
import warnings

btype_sections = {
    "bond": "Bonds",
    "angle": "Angles",
    "dihedral": "Dihedrals",
    "improper": "Impropers",
}


class DCDWriter(DCD.DCDWriter):
    """Write a LAMMPS_ DCD trajectory.

    The units can be set from the constructor with the keyword
    arguments *timeunit* and *lengthunit*. The defaults are "fs" and
    "Angstrom". See :mod:`MDAnalysis.units` for other recognized
    values.
    """

    format = "LAMMPS"
    multiframe = True
    flavor = "LAMMPS"

    def __init__(self, *args, **kwargs):
        self.units = {
            "time": "fs",
            "length": "Angstrom",
        }  # must be instance level
        self.units["time"] = kwargs.pop("timeunit", self.units["time"])
        self.units["length"] = kwargs.pop("lengthunit", self.units["length"])
        for unit_type, unit in self.units.items():
            try:
                if units.unit_types[unit] != unit_type:
                    raise TypeError(
                        f"LAMMPS DCDWriter: wrong unit {unit} for "
                        f"unit type {unit_type}"
                    )
            except KeyError:
                errmsg = f"LAMMPS DCDWriter: unknown unit {unit}"
                raise ValueError(errmsg) from None
        super(DCDWriter, self).__init__(*args, **kwargs)


class DCDReader(DCD.DCDReader):
    """Read a LAMMPS_ DCD trajectory.

    The units can be set from the constructor with the keyword
    arguments *timeunit* and *lengthunit*. The defaults are "fs" and
    "Angstrom", corresponding to LAMMPS `units style`_ "**real**". See
    :mod:`MDAnalysis.units` for other recognized values.

    .. _units style: https://docs.lammps.org/units.html
    """

    format = "LAMMPS"
    flavor = "LAMMPS"

    @store_init_arguments
    def __init__(self, dcdfilename, **kwargs):
        self.units = {
            "time": "fs",
            "length": "Angstrom",
        }  # must be instance level
        self.units["time"] = kwargs.pop("timeunit", self.units["time"])
        self.units["length"] = kwargs.pop("lengthunit", self.units["length"])
        for unit_type, unit in self.units.items():
            try:
                if units.unit_types[unit] != unit_type:
                    raise TypeError(
                        f"LAMMPS DCDReader: wrong unit {unit} for "
                        f"unit type {unit_type}"
                    )
            except KeyError:
                raise ValueError(f"LAMMPS DCDReader: unknown unit {unit}")
        super(DCDReader, self).__init__(dcdfilename, **kwargs)


class DATAReader(base.SingleFrameReaderBase):
    """Reads a single frame of coordinate information from a LAMMPS DATA file.

    .. versionadded:: 0.9.0
    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    """

    format = "DATA"
    units = {"time": None, "length": "Angstrom", "velocity": "Angstrom/fs"}

    @store_init_arguments
    def __init__(self, filename, **kwargs):
        n_atoms = kwargs.pop("n_atoms", None)
        if n_atoms is None:  # this should be done by parsing DATA first
            raise ValueError("DATAReader requires n_atoms keyword")
        self.atom_style = kwargs.pop("atom_style", None)
        super(DATAReader, self).__init__(filename, n_atoms=n_atoms, **kwargs)

    def _read_first_frame(self):
        with DATAParser(self.filename) as p:
            self.ts = p.read_DATA_timestep(
                self.n_atoms, self._Timestep, self._ts_kwargs, self.atom_style
            )

        self.ts.frame = 0
        if self.convert_units:
            self.convert_pos_from_native(self.ts._pos)  # in-place !
            try:
                self.convert_velocities_from_native(
                    self.ts._velocities
                )  # in-place !
            except AttributeError:
                pass


class DATAWriter(base.WriterBase):
    """Write out the current time step as a LAMMPS DATA file.

    This writer supports the sections Atoms, Masses, Velocities, Bonds,
    Angles, Dihedrals, and Impropers. This writer will write the header
    and these sections (if applicable). Atoms section is written in the
    "full" sub-style if charges are available or "molecular" sub-style
    if they are not. Molecule id is set to 0 for all atoms.

    Note
    ----
    This writer assumes "conventional" or "real" LAMMPS units where length
    is measured in Angstroms and velocity is measured in Angstroms per
    femtosecond. To write in different units, specify `lengthunit`

    If atom types are not already positive integers, the user must set them
    to be positive integers, because the writer will not automatically
    assign new types.

    To preserve numerical atom types when writing a selection, the Masses
    section will have entries for each atom type up to the maximum atom type.
    If the universe does not contain atoms of some type in
    {1, ... max(atom_types)}, then the mass for that type will be set to 1.

    In order to write bonds, each selected bond type must be explicitly set to
    an integer >= 1.

    """

    format = "DATA"

    def __init__(self, filename, convert_units=True, **kwargs):
        """Set up a DATAWriter

        Parameters
        ----------
        filename : str
            output filename
        convert_units : bool, optional
            units are converted to the MDAnalysis base format; [``True``]
        """
        self.filename = util.filename(filename, ext="data", keep=True)

        self.convert_units = convert_units

        self.units = {"time": "fs", "length": "Angstrom"}
        self.units["length"] = kwargs.pop("lengthunit", self.units["length"])
        self.units["time"] = kwargs.pop("timeunit", self.units["time"])
        self.units["velocity"] = kwargs.pop(
            "velocityunit", self.units["length"] + "/" + self.units["time"]
        )

    def _write_atoms(self, atoms, data):
        self.f.write("\n")
        self.f.write("Atoms\n")
        self.f.write("\n")

        try:
            charges = atoms.charges
        except (NoDataError, AttributeError):
            has_charges = False
        else:
            has_charges = True

        indices = atoms.indices + 1
        types = atoms.types.astype(np.int32)

        moltags = data.get("molecule_tag", np.zeros(len(atoms), dtype=int))

        if self.convert_units:
            coordinates = self.convert_pos_to_native(
                atoms.positions, inplace=False
            )

        if has_charges:
            for index, moltag, atype, charge, coords in zip(
                indices, moltags, types, charges, coordinates
            ):
                x, y, z = coords
                self.f.write(
                    f"{index:d} {moltag:d} {atype:d} {charge:f}"
                    f" {x:.10f} {y:.10f} {z:.10f}\n"
                )
        else:
            for index, moltag, atype, coords in zip(
                indices, moltags, types, coordinates
            ):
                x, y, z = coords
                self.f.write(
                    f"{index:d} {moltag:d} {atype:d} {x:.10f} {y:.10f} {z:.10f}\n"
                )

    def _write_velocities(self, atoms):
        self.f.write("\n")
        self.f.write("Velocities\n")
        self.f.write("\n")
        indices = atoms.indices + 1
        velocities = self.convert_velocities_to_native(
            atoms.velocities, inplace=False
        )
        for index, vel in zip(indices, velocities):
            self.f.write(
                "{i:d} {x:.10f} {y:.10f} {z:.10f}\n".format(
                    i=index, x=vel[0], y=vel[1], z=vel[2]
                )
            )

    def _write_masses(self, atoms):
        self.f.write("\n")
        self.f.write("Masses\n")
        self.f.write("\n")
        mass_dict = {}
        max_type = max(atoms.types.astype(np.int32))
        for atype in range(1, max_type + 1):
            # search entire universe for mass info, not just writing selection
            masses = set(
                atoms.universe.atoms.select_atoms(
                    "type {:d}".format(atype)
                ).masses
            )
            if len(masses) == 0:
                mass_dict[atype] = 1.0
            else:
                mass_dict[atype] = masses.pop()
            if masses:
                raise ValueError(
                    "LAMMPS DATAWriter: to write data file, "
                    + "atoms with same type must have same mass"
                )
        for atype, mass in mass_dict.items():
            self.f.write("{:d} {:f}\n".format(atype, mass))

    def _write_bonds(self, bonds):
        self.f.write("\n")
        self.f.write("{}\n".format(btype_sections[bonds.btype]))
        self.f.write("\n")
        for bond, i in zip(bonds, range(1, len(bonds) + 1)):
            try:
                self.f.write(
                    "{:d} {:d} ".format(i, int(bond.type))
                    + " ".join((bond.atoms.indices + 1).astype(str))
                    + "\n"
                )
            except TypeError:
                errmsg = (
                    f"LAMMPS DATAWriter: Trying to write bond, but bond "
                    f"type {bond.type} is not numerical."
                )
                raise TypeError(errmsg) from None

    def _write_dimensions(self, dimensions):
        """Convert dimensions to triclinic vectors, convert lengths to native
        units and then write the dimensions section
        """
        if self.convert_units:
            triv = self.convert_pos_to_native(
                mdamath.triclinic_vectors(dimensions), inplace=False
            )
        self.f.write("\n")
        self.f.write("{:f} {:f} xlo xhi\n".format(0.0, triv[0][0]))
        self.f.write("{:f} {:f} ylo yhi\n".format(0.0, triv[1][1]))
        self.f.write("{:f} {:f} zlo zhi\n".format(0.0, triv[2][2]))
        if any([triv[1][0], triv[2][0], triv[2][1]]):
            self.f.write(
                "{xy:f} {xz:f} {yz:f} xy xz yz\n".format(
                    xy=triv[1][0], xz=triv[2][0], yz=triv[2][1]
                )
            )
        self.f.write("\n")

    @requires("types", "masses")
    def write(self, selection, frame=None):
        """Write selection at current trajectory frame to file.

        The sections for Atoms, Masses, Velocities, Bonds, Angles,
        Dihedrals, and Impropers (if these are defined) are
        written. The Atoms section is written in the "full" sub-style
        if charges are available or "molecular" sub-style if they are
        not. Molecule id in atoms section is set to to 0.

        No other sections are written to the DATA file.
        As of this writing, other sections are not parsed into the topology
        by the :class:`DATAReader`.

        Note
        ----
        If the selection includes a partial fragment, then only the bonds,
        angles, etc. whose atoms are contained within the selection will be
        included.

        Parameters
        ----------
        selection : AtomGroup or Universe
            MDAnalysis AtomGroup (selection or Universe.atoms) or also Universe
        frame : int (optional)
            optionally move to frame number `frame`

        """
        u = selection.universe
        if frame is not None:
            u.trajectory[frame]
        else:
            frame = u.trajectory.ts.frame

        # make sure to use atoms (Issue 46)
        atoms = selection.atoms

        # check that types can be converted to ints if they aren't ints already
        try:
            atoms.types.astype(np.int32)
        except ValueError:
            errmsg = (
                "LAMMPS.DATAWriter: atom types must be convertible to "
                "integers"
            )
            raise ValueError(errmsg) from None

        try:
            velocities = atoms.velocities
        except (NoDataError, AttributeError):
            has_velocities = False
        else:
            has_velocities = True

        features = {}
        with util.openany(self.filename, "wt") as self.f:
            self.f.write("LAMMPS data file via MDAnalysis\n")
            self.f.write("\n")
            self.f.write("{:>12d}  atoms\n".format(len(atoms)))

            attrs = [
                ("bond", "bonds"),
                ("angle", "angles"),
                ("dihedral", "dihedrals"),
                ("improper", "impropers"),
            ]

            for btype, attr_name in attrs:
                features[btype] = atoms.__getattribute__(attr_name)
                self.f.write(
                    "{:>12d}  {}\n".format(len(features[btype]), attr_name)
                )
                features[btype] = features[btype].atomgroup_intersection(
                    atoms, strict=True
                )

            self.f.write("\n")
            self.f.write(
                "{:>12d}  atom types\n".format(
                    max(atoms.types.astype(np.int32))
                )
            )

            for btype, attr in features.items():
                self.f.write(
                    "{:>12d}  {} types\n".format(len(attr.types()), btype)
                )

            self._write_dimensions(atoms.dimensions)

            self._write_masses(atoms)
            self._write_atoms(atoms, u.trajectory.ts.data)
            for attr in features.values():
                if attr is None or len(attr) == 0:
                    continue
                self._write_bonds(attr)

            if has_velocities:
                self._write_velocities(atoms)


class DumpReader(base.ReaderBase):
    """Reads the default `LAMMPS dump format
    <https://docs.lammps.org/dump.html>`__

    Supports coordinates in the LAMMPS "unscaled" (x,y,z), "scaled" (xs,ys,zs),
    "unwrapped" (xu,yu,zu) and "scaled_unwrapped" (xsu,ysu,zsu) coordinate
    conventions (see https://docs.lammps.org/dump.html for more details).
    If `lammps_coordinate_convention='auto'` (default),
    one will be guessed. Guessing checks whether the coordinates fit each
    convention in the order "unscaled", "scaled", "unwrapped",
    "scaled_unwrapped" and whichever set of coordinates is detected first will
    be used. If coordinates are given in the scaled coordinate convention
    (xs,ys,zs) or scaled unwrapped coordinate convention (xsu,ysu,zsu) they
    will automatically be converted from their scaled/fractional representation
    to their real values.

    Supports both orthogonal and triclinic simulation box dimensions (for more
    details see https://docs.lammps.org/Howto_triclinic.html). In either case,
    MDAnalysis will always use ``(*A*, *B*, *C*, *alpha*, *beta*, *gamma*)``
    to represent the unit cell. Lengths *A*, *B*, *C* are in the MDAnalysis
    length unit (Å), and angles are in degrees.

    By using the keyword `additional_columns`, you can specify arbitrary data
    to be read. The keyword expects a list of the names of the columns or
    `True` to read all additional columns. The results are saved to
    :attr:`Timestep.data`. For example, if your LAMMPS dump looks like this

    .. code-block::

        ITEM: ATOMS id x y z q l
        1 2.84 8.17 -25 0.00258855 1.1
        2 7.1 8.17 -25 6.91952e-05 1.2

    Then you may parse the additional columns `q` and `l` via:

    .. code-block:: python

        u = mda.Universe('structure.data', 'traj.lammpsdump',
                         additional_columns=['q', 'l'])

    The additional data is then available for each time step via:

    .. code-block:: python

        for ts in u.trajectory:
            charges = ts.data['q'] # Access additional data, sorted by the id
            ls = ts.data['l']
        ...

    Parameters
    ----------
    filename : str
        Filename of LAMMPS dump file
    lammps_coordinate_convention : str (optional) default="auto"
        Convention used in coordinates, can be one of the following according
        to the `LAMMPS documentation <https://docs.lammps.org/dump.html>`__:

         - "auto" - Detect coordinate type from file column header. If auto
           detection is used, the guessing checks whether the coordinates
           fit each convention in the order "unscaled", "scaled", "unwrapped",
           "scaled_unwrapped" and whichever set of coordinates is detected
           first will be used.
         - "scaled" - Coordinates wrapped in box and scaled by box length (see
            note below), i.e., xs, ys, zs
         - "scaled_unwrapped" - Coordinates unwrapped and scaled by box length,
           (see note below) i.e., xsu, ysu, zsu
         - "unscaled" - Coordinates wrapped in box, i.e., x, y, z
         - "unwrapped" - Coordinates unwrapped, i.e., xu, yu, zu

        If coordinates are given in the scaled coordinate convention (xs,ys,zs)
        or scaled unwrapped coordinate convention (xsu,ysu,zsu) they will
        automatically be converted from their scaled/fractional representation
        to their real values.
    unwrap_images : bool (optional) default=False
        If `True` and the dump file contains image flags, the coordinates
        will be unwrapped. See `read_data
        <https://docs.lammps.org/read_data.html>`__  in the lammps
        documentation for more information.
    **kwargs
       Other keyword arguments used in
       :class:`~MDAnalysis.coordinates.base.ReaderBase`


    .. versionchanged:: 2.8.0
       Reading of arbitrary, additional columns is now supported.
       (Issue `#3504 <https://github.com/MDAnalysis/mdanalysis/issues/3504>`__)
    .. versionchanged:: 2.4.0
       Now imports velocities and forces, translates the box to the origin,
       and optionally unwraps trajectories with image flags upon loading.
    .. versionchanged:: 2.2.0
       Triclinic simulation boxes are supported.
       (Issue `#3383 <https://github.com/MDAnalysis/mdanalysis/issues/3383>`__)
    .. versionchanged:: 2.0.0
       Now parses coordinates in multiple lammps conventions (x,xs,xu,xsu)
    .. versionadded:: 0.19.0
    """

    format = "LAMMPSDUMP"
    _conventions = [
        "auto",
        "unscaled",
        "scaled",
        "unwrapped",
        "scaled_unwrapped",
    ]

    _coordtype_column_names = {
        "unscaled": ["x", "y", "z"],
        "scaled": ["xs", "ys", "zs"],
        "unwrapped": ["xu", "yu", "zu"],
        "scaled_unwrapped": ["xsu", "ysu", "zsu"],
    }

    _parsable_columns = ["id", "vx", "vy", "vz", "fx", "fy", "fz"]
    for key in _coordtype_column_names.keys():
        _parsable_columns += _coordtype_column_names[key]

    @store_init_arguments
    def __init__(
        self,
        filename,
        lammps_coordinate_convention="auto",
        unwrap_images=False,
        additional_columns=None,
        **kwargs,
    ):
        super(DumpReader, self).__init__(filename, **kwargs)

        root, ext = os.path.splitext(self.filename)
        if lammps_coordinate_convention in self._conventions:
            self.lammps_coordinate_convention = lammps_coordinate_convention
        else:
            option_string = "'" + "', '".join(self._conventions) + "'"
            raise ValueError(
                "lammps_coordinate_convention="
                f"'{lammps_coordinate_convention}'"
                " is not a valid option. "
                f"Please choose one of {option_string}"
            )

        self._unwrap = unwrap_images

        if (
            util.iterable(additional_columns)
            or additional_columns is None
            or additional_columns is True
        ):
            self._additional_columns = additional_columns
        else:
            raise ValueError(
                f"additional_columns={additional_columns} "
                "is not a valid option. Please provide an "
                "iterable containing the additional"
                "column headers."
            )

        self._cache = {}

        self._reopen()

        self._read_next_timestep()

    def _reopen(self):
        self.close()
        self._file = util.anyopen(self.filename)
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        self.ts.frame = -1

    @property
    @cached("n_atoms")
    def n_atoms(self):
        with util.anyopen(self.filename) as f:
            f.readline()
            f.readline()
            f.readline()
            n_atoms = int(f.readline())
        return n_atoms

    @property
    @cached("n_frames")
    def n_frames(self):
        # 2(timestep) + 2(natoms info) + 4(box info) + 1(atom header) + n_atoms
        lines_per_frame = self.n_atoms + 9
        offsets = []
        counter = 0
        with util.anyopen(self.filename) as f:
            line = True
            while line:
                if not counter % lines_per_frame:
                    offsets.append(f.tell())
                line = f.readline()
                counter += 1
        self._offsets = offsets[:-1]  # last is EOF
        return len(self._offsets)

    def close(self):
        if hasattr(self, "_file"):
            self._file.close()

    def _read_frame(self, frame):
        self._file.seek(self._offsets[frame])
        self.ts.frame = frame - 1  # gets +1'd in next

        return self._read_next_timestep()

    def _read_next_timestep(self):
        f = self._file
        ts = self.ts
        ts.frame += 1
        if ts.frame >= len(self):
            raise EOFError

        f.readline()  # ITEM TIMESTEP
        step_num = int(f.readline())
        ts.data["step"] = step_num
        ts.data["time"] = step_num * ts.dt

        f.readline()  # ITEM NUMBER OF ATOMS
        n_atoms = int(f.readline())
        if n_atoms != self.n_atoms:
            raise ValueError(
                "Number of atoms in trajectory changed "
                "this is not supported in MDAnalysis"
            )

        triclinic = len(f.readline().split()) == 9  # ITEM BOX BOUNDS
        if triclinic:
            xlo_bound, xhi_bound, xy = map(float, f.readline().split())
            ylo_bound, yhi_bound, xz = map(float, f.readline().split())
            zlo, zhi, yz = map(float, f.readline().split())

            # converts orthogonal bounding box to the conventional format,
            # see https://docs.lammps.org/Howto_triclinic.html
            xlo = xlo_bound - min(0.0, xy, xz, xy + xz)
            xhi = xhi_bound - max(0.0, xy, xz, xy + xz)
            ylo = ylo_bound - min(0.0, yz)
            yhi = yhi_bound - max(0.0, yz)

            box = np.zeros((3, 3), dtype=np.float64)
            box[0] = xhi - xlo, 0.0, 0.0
            box[1] = xy, yhi - ylo, 0.0
            box[2] = xz, yz, zhi - zlo

            xlen, ylen, zlen, alpha, beta, gamma = mdamath.triclinic_box(*box)
        else:
            xlo, xhi = map(float, f.readline().split())
            ylo, yhi = map(float, f.readline().split())
            zlo, zhi = map(float, f.readline().split())
            xlen = xhi - xlo
            ylen = yhi - ylo
            zlen = zhi - zlo
            alpha = beta = gamma = 90.0
        ts.dimensions = xlen, ylen, zlen, alpha, beta, gamma

        indices = np.zeros(self.n_atoms, dtype=int)

        atomline = f.readline()  # ITEM ATOMS etc
        attrs = atomline.split()[2:]  # attributes on coordinate line
        attr_to_col_ix = {x: i for i, x in enumerate(attrs)}
        convention_to_col_ix = {}
        for cv_name, cv_col_names in self._coordtype_column_names.items():
            try:
                convention_to_col_ix[cv_name] = [
                    attr_to_col_ix[x] for x in cv_col_names
                ]
            except KeyError:
                pass

        if self._unwrap:
            try:
                image_cols = [attr_to_col_ix[x] for x in ["ix", "iy", "iz"]]
            except:
                raise ValueError(
                    "Trajectory must have image flag in order " "to unwrap."
                )

        self._has_vels = all(x in attr_to_col_ix for x in ["vx", "vy", "vz"])
        if self._has_vels:
            ts.has_velocities = True
            vel_cols = [attr_to_col_ix[x] for x in ["vx", "vy", "vz"]]
        self._has_forces = all(x in attr_to_col_ix for x in ["fx", "fy", "fz"])
        if self._has_forces:
            ts.has_forces = True
            force_cols = [attr_to_col_ix[x] for x in ["fx", "fy", "fz"]]

        # this should only trigger on first read of "ATOM" card, after which it
        # is fixed to the guessed value. Auto proceeds unscaled -> scaled
        # -> unwrapped -> scaled_unwrapped
        if self.lammps_coordinate_convention == "auto":
            try:
                # this will automatically select in order of priority
                # unscaled, scaled, unwrapped, scaled_unwrapped
                self.lammps_coordinate_convention = list(convention_to_col_ix)[
                    0
                ]
            except IndexError:
                raise ValueError("No coordinate information detected")
        elif not self.lammps_coordinate_convention in convention_to_col_ix:
            raise ValueError(
                "No coordinates following convention "
                f"{self.lammps_coordinate_convention} found in timestep"
            )

        coord_cols = convention_to_col_ix[self.lammps_coordinate_convention]
        if self._unwrap:
            coord_cols.extend(image_cols)

        ids = "id" in attr_to_col_ix

        # Create the data arrays for additional attributes which will be saved
        # under ts.data
        if self._additional_columns is True:
            # Parse every column that is not already parsed
            # elsewhere (total \ parsable)
            additional_keys = set(attrs).difference(self._parsable_columns)
        elif self._additional_columns:
            if not all([key in attrs for key in self._additional_columns]):
                warnings.warn(
                    "Some of the additional columns are not present "
                    "in the file, they will be ignored"
                )
            additional_keys = [
                key for key in self._additional_columns if key in attrs
            ]
        else:
            additional_keys = []
        for key in additional_keys:
            ts.data[key] = np.empty(self.n_atoms)

        # Parse all the atoms
        for i in range(self.n_atoms):
            fields = f.readline().split()
            if ids:
                indices[i] = fields[attr_to_col_ix["id"]]
            coords = np.array(
                [fields[dim] for dim in coord_cols], dtype=np.float32
            )

            if self._unwrap:
                images = coords[3:]
                coords = coords[:3]
                coords += images * ts.dimensions[:3]
            else:
                coords = coords[:3]
            ts.positions[i] = coords

            if self._has_vels:
                ts.velocities[i] = [fields[dim] for dim in vel_cols]
            if self._has_forces:
                ts.forces[i] = [fields[dim] for dim in force_cols]

            # Collect additional cols
            for attribute_key in additional_keys:
                ts.data[attribute_key][i] = fields[
                    attr_to_col_ix[attribute_key]
                ]

        order = np.argsort(indices)
        ts.positions = ts.positions[order]
        if self._has_vels:
            ts.velocities = ts.velocities[order]
        if self._has_forces:
            ts.forces = ts.forces[order]

        # Also need to sort the additional keys
        for attribute_key in additional_keys:
            ts.data[attribute_key] = ts.data[attribute_key][order]

        if self.lammps_coordinate_convention.startswith("scaled"):
            # if coordinates are given in scaled format, undo that
            ts.positions = distances.transform_StoR(
                ts.positions, ts.dimensions
            )
        # Transform to origin after transformation of scaled variables
        ts.positions -= np.array([xlo, ylo, zlo])[None, :]

        return ts
